
//		BOTTOMUP command:


#ifndef BOTTOMUP_VER
#define BOTTOMUP_VER					1.0.0.1



const wchar_t BottomUpHelpText[] = L"Visit subdirectories, deepest first, and run a command in each.  \n\
\n\
BOTTOMUP /B /H /I /N /P /Q /S:dir command\r\n\
\n\
\t/B\tbare directory display\n\
\t/H\thidden\n\
\t/I\tignore exit codes\n\
\t/N\tno junctions\n\
\t/P\tprompt\n\
\t/Q\tquietly\n\
\t/S:\tstarting directory\n\
\n\
This command will set the current working directory to each subdirectory\n\
in turn, deepest first.  The specified command will be run in each.  Double  \n\
any percent signs, and escape or strong-quote any problem characters.  You\n\
must include a command; use REM if you want to do nothing.\n";




int GetYNorA( BOOL ShowPrompt, BOOL AllowAll )
{
	unsigned int ch = 0x00;

	EatKeystrokes();

	if ( ShowPrompt ) {
		if ( AllowAll )
			_cputws( L" (Y/N/A)? " );
		else
			_cputws( L" (Y/N)? " );
	}

	do {
		if ( QueryKeyWaiting() ) {

			ch = GetKeystroke( 0x0030 );

			ch = towupper( ch );

			if ( ch == 'Y' ) {
				_putwch( ch );
				_putwch( '\n' );
				return 1;
			}
			else if ( ( ch == 'N' ) || ( ch == 0x1b ) ) {
				_putwch( 'N' );
				_putwch( '\n' );
				return 0;
			}
			else if ( AllowAll && ( ( ch == 'A' ) || ( ch == 'R' ) ) ) {
				_putwch( ch );
				_putwch( '\n' );
				return -1;
			}
			else
				honk();
		}
		else
			tty_yield( 0 );

	} while ( TRUE );

	return 0;
}



DLLExports int bottomup( LPTSTR inString )
{
	BOOL Ignore					= FALSE;
	BOOL Prompt					= FALSE;
	BOOL Quiet					= FALSE;
	BOOL Bare					= FALSE;
	BOOL DoIt					= TRUE;


	wchar_t UserCommand[MAX_ARG_LEN]		= L"";
	wchar_t OriginalCWD[MAX_FILENAME_LEN]	= L"";
	wchar_t StartPath[MAX_FILENAME_LEN]		= L"";
	wchar_t Wildspec[MAX_FILENAME_LEN]		= L"";
	wchar_t CmdTemp[MAX_ARG_LEN]			= L"";
	wchar_t Mask[MAX_ARG_LEN]				= L"*";

	int Recurse								= FH_FLAG_RECURSE | FH_FLAG_RECURSE_BOTTOM_UP | FH_FLAG_NO_HIDDEN_DIRS | FH_FLAG_SUPPRESS_FNF_ERROR;
	DWORD AttribMask						= FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_HIDDEN;
	DWORD AttribValues						= FILE_ATTRIBUTE_DIRECTORY;

	FH_FilenamesLinkedList DirList			= { NULL, NULL, 0 };
	FH_FilenameItem *This					= NULL;

	int rv			= 0;



	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE | PARSE_ONE_ARG_KLUDGE | PARSE_ONE_ARG_KEEP_QUOTES );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcsncmp( arg[i], L"/?", 2 ) )
				return ShowCmdHelp( BottomUpHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( BOTTOMUP_VER ) );
			else if ( !_wcsicmp( arg[i], L"/B" ) )
				Bare = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/D", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing namespec: \"%s\"\r\n", arg[i] );
					return 1;
				}
				
				wcscpy_s( Mask, MAX_ARG_LEN, arg[i] + j );
			}
			else if ( !_wcsicmp( arg[i], L"/H" ) ) {
				AttribMask &= ( ~FILE_ATTRIBUTE_HIDDEN );
				Recurse &= ( ~FH_FLAG_NO_HIDDEN_DIRS );
			}
			else if ( !_wcsicmp( arg[i], L"/I" ) )
				Ignore = TRUE;
			else if ( !_wcsicmp( arg[i], L"/N" ) ) {
				AttribMask |= FILE_ATTRIBUTE_REPARSE_POINT;
				AttribValues &= ( ~FILE_ATTRIBUTE_REPARSE_POINT );
				Recurse |= FH_FLAG_NO_JUNCTIONS;
			}
			else if ( !_wcsicmp( arg[i], L"/P" ) )
				Prompt = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Q" ) )
				Quiet = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/S", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing directory name: \"%s\"\r\n", arg[i] );
					return 1;
				}

				StartPath[0] = '\0';
				FH_Truename( arg[i] + j, StartPath );

				if ( StartPath[0] == '\0' ) {
					_wcsupr_s( arg[i], MAX_ARG_LEN );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad or missing directory name: \"%s\"\r\n", arg[i] );
					return 1;
				}

				if ( !QueryIsDirectory( StartPath ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Not a directory: \"%s\"\r\n", StartPath );
					return 1;
				}
			}
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}
			else
				wcscpy_s( UserCommand, MAX_ARG_LEN, arg[i] );
		}
	}


	if ( UserCommand[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing command\r\n\r\n" );
		return 1;
	}

	if ( Prompt || Bare )
		Quiet = FALSE;



	QueryTrueName( L".", OriginalCWD );

	if ( StartPath[0] == '\0' )
		wcscpy_s( StartPath, MAX_FILENAME_LEN, OriginalCWD );

	wcscpy_s( Wildspec, MAX_FILENAME_LEN, StartPath );
	AppendFilename( Wildspec, Mask, MAX_FILENAME_LEN );


	HandleMatchingFiles( Wildspec, 0, &DirList, FH_AddFileToLinkedList, Recurse, AttribMask, AttribValues, NULL, 0, UINT_MAX );
	FH_AddFileToLinkedList( StartPath, 0, &DirList );


	This = DirList.First;

	while ( This ) {

		if ( ChangeDirectory( &(This->Filename) ) ) {
			GetErrorText( GetLastError(), CmdTemp );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error moving to directory \"%s\" :\r\n", &(This->Filename) );
			Qprintf( ERRHANDLE, L"   %s\r\n", CmdTemp );

			if ( !Ignore ) {
				rv = 2;
				break;
			}
		}
		else {
			//		changed to the directory successfully:

			if ( !Quiet ) {
				if ( !Bare )
					QPuts( L"\r\nBOTTOMUP: " );

				QPuts( &(This->Filename) );
				QPuts( L"\r\n" );
			}

			wcscpy_s( CmdTemp, MAX_ARG_LEN, UserCommand );

			if ( Prompt ) {
				_cwprintf( L"    \u00BB %s", UserCommand );

				int yn = GetYNorA( TRUE, TRUE );

				if ( yn < 0 )
					DoIt = TRUE, Prompt = FALSE;
				else if ( yn )
					DoIt = TRUE;
				else
					DoIt = FALSE;
			}

			if ( DoIt )
				rv = Command( CmdTemp, 0 );

			if ( rv && !Ignore )
				break;

		}

		This = This->Next;
	}

	ChangeDirectory( OriginalCWD );
	FH_FreeLinkedList( &(DirList.First) );
	return rv;
}


#endif	//	BOTTOMUP_VER


/*

	1.0.0.0		2018-05-10

	1.0.0.1		2022-03-01
				Help text is now displayed via ShowCmdHelp().

*/
