
//		FILEINFO command, and related structures and functions:


#ifndef FILEINFO_VER
#define FILEINFO_VER					1.0.9.3


#define FILEINFO_OMIT_VERHEADER			0x0001
#define FILEINFO_OMIT_SIZE				0x0002
#define FILEINFO_OMIT_TIMESTAMPS		0x0004
#define FILEINFO_OMIT_ATTRIBUTES		0x0008
#define FILEINFO_OMIT_CHECKSUMS			0x0010
#define FILEINFO_OMIT_4DESCRIPTION		0x0020
#define FILEINFO_OMIT_VERINFO			0x0040
#define FILEINFO_OMIT_SUMMARY_INFO		0x0080
#define FILEINFO_OMIT_IMAGE_INFO		0x0100
#define FILEINFO_OMIT_WIN_PROPS			0x0200
#define FILEINFO_OMIT_EXE_INFO			0x0400
#define FILEINFO_OMIT_TEXT_ENCODING		0x0800
#define FILEINFO_OMIT_LNK_TARGET		0x1000

#define FILEINFO_TIME_CREATE			0x0001
#define FILEINFO_TIME_MODIFY			0x0002
#define FILEINFO_TIME_ACCESS			0x0004
#define FILEINFO_TIME_ALL				0x0007

#define FILEINFO_HASH_CRC32				0x0001
#define FILEINFO_HASH_CKSUM				0x0002
#define FILEINFO_HASH_CKSUMX			0x0004
#define FILEINFO_HASH_MD5				0x0010
#define FILEINFO_HASH_SHA1				0x0020
#define FILEINFO_HASH_SHA256			0x0040
#define FILEINFO_HASH_SHA384			0x0080
#define FILEINFO_HASH_SHA512			0x0100


#include <propkey.h>
#include <propvarutil.h>

#define EBCDIC_SUPPORT
#include "FileHandler.cpp"
#include "conlist.cpp"
#include "LnkTarget.cpp"



enum _FileInfo_TimeDisplay { Default = 0, Smart, UTC };


struct _FileInfoArgs {
	unsigned int Count;
	_FileInfo_TimeDisplay TimeDisplay;
	unsigned int TimeStamps;
	unsigned int Hash;
	LPCTSTR TextFileTypes;
};



const wchar_t FileInfoHelpText[] = L"Display information about files.\n\
\n\
FILEINFO /A: /C: /D /N: /O: /P /S /T: file...\r\n\
\n\
\t/A:\tattributes\n\
\t/C:\tchecksums/hashes    /C? for more info  \n\
\t/D\tno highlight\n\
\t/N:\tdisable features    /N? for more info\n\
\t/O:\tomit features       /O? for more info\n\
\t/P\tpage output\n\
\t/S\tsearch in subdirectories\n\
\t/T:\ttime display mode   /T? for more info\n\
\n\
If no filespecs are given, * is the default.  \n";


const wchar_t FileInfoOptCHelpText[]		= L"FILEINFO /C:  \u2014 checksum/hash algorithms\n\
\n\
\tCRC32\t\tCKSUM\t\tMD5\n\
\tSHA1\t\tSHA256\t\tSHA384\n\
\tSHA512\n\
\n\
You can specify more than one, separated by commas.\n\
If no /C: is given, the default is /C:CRC32,MD5.\n";


#ifdef IMGINFO_VER

const wchar_t FileInfoOptOHelpText[]		= L"FILEINFO /O:  \u2014 omit features\n\
\n\
\tA\tattributes\n\
\tC\tchecksums\n\
\tD\tfile descriptions\n\
\tE\ttext file encoding\n\
\tH\tfile version header\n\
\tI\timage file info\n\
\tL\tlink targets\n\
\tP\tWindows properties\n\
\tS\tdocument summary info\n\
\tT\ttime stamps\n\
\tV\tversion info\n\
\tX\texecutable info\n\
\tZ\tsizes\n";

#else	//	IMGINFO_VER

const wchar_t FileInfoOptOHelpText[]		= L"FILEINFO /O:  \u2014 omit features\n\
\n\
\tA\tattributes\n\
\tC\tchecksums\n\
\tD\tfile descriptions\n\
\tE\ttext file encoding\n\
\tH\tfile version header\n\
\tL\tlink targets\n\
\tP\tWindows properties\n\
\tS\tdocument summary info\n\
\tT\ttime stamps\n\
\tV\tversion info\n\
\tX\texecutable info\n\
\tZ\tsizes\n";

#endif	//	IMGINFO_VER


const wchar_t FileInfoOptNHelpText[]		= L"FILEINFO /N:  \u2014 disable features\n\
\n\
\tC\tno highlight\n\
\tD\tdo not search into hidden directories\n\
\tF\tsuppress file-not-found error\n\
\tJ\tdo not search into junctions\n\
\tZ\tdo not search into system directories\n";


const wchar_t FileInfoOptTHelpText[]		= L"FILEINFO /T:  \u2014 time display mode\n\
\n\
\tA\tshow last access time\n\
\tC\tshow file creation time\n\
\tM\tshow last modification time\n\
\tD\tdefault:  localize times as TCC and CMD do\n\
\tL\tlocal times, taking DST time changes into account\n\
\tU\tCoordinated Universal Time\n\
\n\
The default is to show all three stamps, localized as TCC\n\
and CMD do.\n";



const wchar_t FI_DowNames[8][4]			= { L"Sun", L"Mon", L"Tue", L"Wed", L"Thu", L"Fri", L"Sat", L"XXX" };


const wchar_t FI_ExeTypeStr[19][25]		= { L"Unknown", L"MS-DOS", L".PIF file", L"Windows 16-bit", L"Windows 3.x VxD", L"OS/2",
											L"Windows 32-bit", L"Windows 32-bit (console)", L"Posix", L"Windows 64-bit", L"Windows 64-bit (console)",
											L"EFI application", L"EFI boot service driver", L"EFI runtime driver", L"EFI ROM", L"XBox", 
											L"Windows boot application", L"32-bit .DLL file", L"64-bit .DLL file" };

const wchar_t FI_DefaultTextFileTypes[]	= L".txt;.log;.cpp;.html";

const wchar_t FI_TextFileTypesVar[]		= L"TextFileTypes";


const int FI_NumHashes					= 14;

const wchar_t FI_HashName[FI_NumHashes][8]		= { L"CRC32", L"CRC-32", L"CKSUM", L"CKSUMX", L"MD5", L"MD-5", L"SHA1", L"SHA-1", L"SHA256", L"SHA-256",
													L"SHA384", L"SHA-384", L"SHA512", L"SHA-512" };

const unsigned int FI_HashVal[FI_NumHashes]		= { FILEINFO_HASH_CRC32, FILEINFO_HASH_CRC32, FILEINFO_HASH_CKSUM, FILEINFO_HASH_CKSUMX, FILEINFO_HASH_MD5, FILEINFO_HASH_MD5,
													FILEINFO_HASH_SHA1, FILEINFO_HASH_SHA1, FILEINFO_HASH_SHA256, FILEINFO_HASH_SHA256,
													FILEINFO_HASH_SHA384, FILEINFO_HASH_SHA384, FILEINFO_HASH_SHA512, FILEINFO_HASH_SHA512 };



#ifdef IMGINFO_VER

#ifndef EPOCH
#define EPOCH			1601
#define ULTIMA			9999
#endif	//	EPOCH

const int FI_MonthLength[13][2]	= { {0, 0}, {31, 31}, {28, 29}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31}, {31, 31}, {30, 30}, {31, 31}, {30, 30}, {31, 31} };

const int FI_Kalends[13][2]		= { {0, 0}, {0, 0}, {31, 31}, {59, 60}, {90, 91}, {120, 121}, {151, 152}, {181, 182}, {212, 213}, {243, 244}, {273, 274}, {304, 305}, {334, 335} };

#endif	//	IMGINFO_VER


#include "FileChecksums.cpp"



int InsertCommas( LPTSTR Buffer, unsigned int BufferLen )
{
	if ( !Buffer )
		return -1;

	if ( wcslen( Buffer ) > ( MAX_ARG_LEN - 10 ) )
		return 1;

	wchar_t temp[MAX_ARG_LEN]	= L"%@comma[";
	wcscat_s( temp, MAX_ARG_LEN, Buffer );
	wcscat_s( temp, MAX_ARG_LEN, L"]" );

	if ( ExpandVariables( temp, 1 ) )
		return 2;

	if ( wcslen( temp ) >= BufferLen )
		return 3;

	wcscpy_s( Buffer, BufferLen, temp );
	return 0;
}



int DisplayFileSizeCosmetic( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	unsigned __int64 FileSize		= 0;

	wchar_t Buffer[SHORT_BUF_LEN]	= L"";


	FileSize = QueryFileSize( (LPTSTR) Filename, 0 );

	wsprintf( Buffer, L"%I64u", FileSize );
	InsertCommas( Buffer, SHORT_BUF_LEN );

	if ( FileSize == 1 )
		QPuts( L"    1 byte     " );
	else
		Printf( L"    %s bytes     ", Buffer );

	if ( FileSize >= ( 1024 * 1024 * 1024 ) ) {
		swprintf( Buffer, SHORT_BUF_LEN, L"%.2f", (double) FileSize / 1073741824.0 );
		InsertCommas( Buffer, SHORT_BUF_LEN );
		Printf( L"%s gigabytes", Buffer );
	}
	else if ( FileSize >= ( 1024 * 1024 ) ) {
		swprintf( Buffer, SHORT_BUF_LEN, L"%.2f", (double) FileSize / 1048576.0 );
		InsertCommas( Buffer, SHORT_BUF_LEN );
		Printf( L"%s megabytes", Buffer );
	}
	else if ( FileSize >= 1024 ) {
		swprintf( Buffer, SHORT_BUF_LEN, L"%.2f", (double) FileSize / 1024.0 );
		InsertCommas( Buffer, SHORT_BUF_LEN );
		Printf( L"%s kilobytes", Buffer );
	}

	return CL_EndLine( 2 );
}



int FI_ParseExeHeader( LPCTSTR Filename, WORD *MachineType, WORD *Subsystem )
{
	if ( !Filename || !MachineType || !Subsystem )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	const int BufferSize					= 8192;
	BYTE Buffer[BufferSize];

	HANDLE FileHandle						= NULL;
	DWORD BytesRead							= 0;

	_IMAGE_DOS_HEADER *DosHeader			= NULL;
	_IMAGE_FILE_HEADER *ImageHeader			= NULL;
	_IMAGE_OPTIONAL_HEADER *OptHeader		= NULL;
	_IMAGE_OPTIONAL_HEADER64 *OptHeader64	= NULL;
	_IMAGE_VXD_HEADER *VxdHeader			= NULL;

	WORD *WordPointer						= NULL;
	DWORD *DWordPointer						= NULL;

	int rv									= 1;



	*MachineType	= 0;
	*Subsystem		= 0;


	memset ( &Buffer, 0x00, BufferSize );


	FileHandle = CreateFile( Filename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );

	if ( FileHandle != INVALID_HANDLE_VALUE ) {

		if ( ReadFile( FileHandle, Buffer, BufferSize, &BytesRead, NULL ) ) {
			if ( BytesRead >= 1024 ) {

				DosHeader = (_IMAGE_DOS_HEADER*) Buffer;

				if ( ( DosHeader->e_magic == 0x5a4d ) || ( DosHeader->e_magic == 0x4d5a ) ) {
					//	wprintf( L"   * Found EXE header.\n" );

					if ( DosHeader->e_lfanew && ( ( DosHeader->e_lfanew + sizeof( _IMAGE_FILE_HEADER ) + sizeof( _IMAGE_OPTIONAL_HEADER ) ) < BytesRead ) ) {
						DWordPointer = (DWORD*) ( Buffer + DosHeader->e_lfanew );
						//	wprintf( L"   * Signature = 0x%08X\n", *DWordPointer );

						if ( *DWordPointer == IMAGE_NT_SIGNATURE ) {

							ImageHeader = (_IMAGE_FILE_HEADER*) ( Buffer + DosHeader->e_lfanew + 4 );
							*MachineType = ImageHeader->Machine;

							if ( ImageHeader->SizeOfOptionalHeader >= 20 ) {

								WordPointer = (WORD*) ( (BYTE*) ImageHeader + sizeof( _IMAGE_FILE_HEADER ) );
								//	wprintf( L"   * Magic number = 0x%04X\n", *WordPointer );

								if ( ( *WordPointer == IMAGE_NT_OPTIONAL_HDR32_MAGIC ) && ( ImageHeader->SizeOfOptionalHeader >= sizeof( _IMAGE_OPTIONAL_HEADER ) ) ) {
									OptHeader = (_IMAGE_OPTIONAL_HEADER*) WordPointer;
									*Subsystem = OptHeader->Subsystem;
									rv = 0;
								}
								else if ( ( *WordPointer == IMAGE_NT_OPTIONAL_HDR64_MAGIC ) && ( ImageHeader->SizeOfOptionalHeader >= sizeof( _IMAGE_OPTIONAL_HEADER64 ) ) ) {
									OptHeader64 = (_IMAGE_OPTIONAL_HEADER64*) WordPointer;
									*Subsystem = OptHeader64->Subsystem;
									rv = 0;
								}

							}
						}
						else if ( ( *DWordPointer & 0x0000ffff ) == IMAGE_VXD_SIGNATURE ) {
							//	wprintf( L"   * File might be a VXD!\n" );

							VxdHeader = (_IMAGE_VXD_HEADER*) ( Buffer + DosHeader->e_lfanew );

							if ( ( VxdHeader->e32_border == 0 ) && ( VxdHeader->e32_worder == 0 ) && ( VxdHeader->e32_level == 0 ) && ( VxdHeader->e32_cpu >= 2 ) && ( VxdHeader->e32_cpu <= 4 ) && ( VxdHeader->e32_os == 4 ) ) {
								*MachineType = IMAGE_FILE_MACHINE_I386;
								*Subsystem = IMAGE_VXD_SIGNATURE;
								rv = 0;
							}
						}
					}
				}
			}
		}

		CloseHandle( FileHandle );
	}

	return rv;
}



unsigned int FI_NonBlanks( LPCTSTR String )
{
	if ( !String )
		return 0;


	unsigned int k		= 0;


	for ( const wchar_t *ch = String; *ch; ch++ )
		if ( !iswspace( *ch ) )
			k++;

	return k;
}



int FI_GetFileVerString( void *Buffer, LPTSTR Field, LPTSTR outString, size_t BufferLen )
{
	if ( !Buffer || !Field || !outString )
		return -1;


	struct LangCodePagePair {
		WORD Language;
		WORD CodePage;
	};

	LangCodePagePair *Translations		= NULL;
	unsigned int TranslationsLen		= 0;
	unsigned int NumTranslations		= 0;

	wchar_t FieldName[SHORT_BUF_LEN]	= L"";
	void *Item							= NULL;
	UINT ItemLen						= 0;


	VerQueryValue( Buffer, L"\\VarFileInfo\\Translation", (void**) &Translations, &TranslationsLen );
	if ( !TranslationsLen )
		return 1;

	NumTranslations = TranslationsLen / sizeof( LangCodePagePair );


	for ( unsigned int i = 0; i < NumTranslations; i++ ) {
		wsprintf( FieldName, L"\\StringFileInfo\\%04X%04X\\%s", Translations[i].Language, Translations[i].CodePage, Field );
		if ( VerQueryValue( Buffer, FieldName, &Item, &ItemLen ) ) {
			wcscpy_s( outString, BufferLen, (LPTSTR) Item );
			return 0;
		}
	}

	return 2;
}



int DisplayFileVerHeader( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;



	wchar_t FileDescription[SHORT_BUF_LEN]		= L"";
	wchar_t FileVersionString[SHORT_BUF_LEN]	= L"";
	wchar_t CopyrightString[SHORT_BUF_LEN]		= L"";
	
	DWORD VerInfoSize					= 0;
	void *Buffer						= NULL;
	BOOL Anything						= FALSE;



	if ( !IsFile( Filename ) )
		return ERROR_FILE_NOT_FOUND;


	VerInfoSize = GetFileVersionInfoSize( Filename, NULL );
	if ( !VerInfoSize )
		return ERROR_MOD_NOT_FOUND;

	Buffer = malloc( VerInfoSize );
	if ( !Buffer )
		return ERROR_NOT_ENOUGH_MEMORY;

	if ( GetFileVersionInfo( Filename, 0, VerInfoSize, Buffer ) ) {
		FI_GetFileVerString( Buffer, L"FileDescription", FileDescription, SHORT_BUF_LEN );
		FI_GetFileVerString( Buffer, L"FileVersion", FileVersionString, SHORT_BUF_LEN );
		FI_GetFileVerString( Buffer, L"LegalCopyright", CopyrightString, SHORT_BUF_LEN );
	}

	if ( FI_NonBlanks( FileDescription ) && FI_NonBlanks( FileVersionString ) ) {
		Printf( L"    \"%s\" \u2014 %s", FileDescription, FileVersionString );
		Anything = TRUE;

		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}
	else if ( FI_NonBlanks( FileDescription ) ) {
		Printf( L"    \"%s\" \u2014 (no version)", FileDescription );
		Anything = TRUE;

		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}
	else if ( FI_NonBlanks( FileVersionString ) ) {
		Printf( L"    (No description) \u2014 %s", FileVersionString );
		Anything = TRUE;

		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}

	if ( FI_NonBlanks( CopyrightString ) ) {
		Printf( L"    %s", CopyrightString );
		Anything = TRUE;

		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;
	}

	if ( Anything )
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;

	free( Buffer );
	return 0;

UserAbort:

	free( Buffer );
	return USER_ABORT;
}



int FI_GetDirectoryInfo( LPCTSTR Filename, WIN32_FIND_DATA *FileInfo )
{
	if ( !Filename || !FileInfo )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	HANDLE SearchHandle			= NULL;
	WIN32_FIND_DATA Found;
	LPTSTR NamePart				= FH_FilenamePart( Filename, NULL );

	int rv						= 0;


	if ( !NamePart )
		return -1;

	if ( wcschr( NamePart, '*' ) || wcschr( NamePart, '?' ) )
		return -1;


	memset( FileInfo, 0x00, sizeof( WIN32_FIND_DATA ) );
	memset( &Found, 0x00, sizeof( WIN32_FIND_DATA ) );


	SearchHandle = FindFirstFile( Filename, &Found );
	if ( SearchHandle != INVALID_HANDLE_VALUE ) {
		if ( !_wcsicmp( Found.cFileName, NamePart ) ) {
			*FileInfo = Found;
			rv = 1;
		}

		FindClose( SearchHandle );
	}

	return rv;
}



int FI_ShowStamp( FILETIME Timestamp, LPTSTR Preface, _FileInfo_TimeDisplay How )
{
	FILETIME LocalTime;
	SYSTEMTIME UtcTime;
	SYSTEMTIME Display;
	LPCTSTR Dow				= FI_DowNames[0];


	if ( Preface )
		QPuts( Preface );


	if ( ( Timestamp.dwHighDateTime | Timestamp.dwLowDateTime ) == 0 ) {
		Printf( L"(none)" );
		return CL_EndLine();
	}


	if ( How == UTC )
		FileTimeToSystemTime( &Timestamp, &Display );
	else if ( How == Smart ) {
		FileTimeToSystemTime( &Timestamp, &UtcTime );
		SystemTimeToTzSpecificLocalTime( NULL, &UtcTime, &Display );
	}
	else {
		FileTimeToLocalFileTime( &Timestamp, &LocalTime );
		FileTimeToSystemTime( &LocalTime, &Display );
	}

	if ( Display.wDayOfWeek < 7 )
		Dow = FI_DowNames[Display.wDayOfWeek];
	else
		Dow = FI_DowNames[7];

	Printf( L"%s %04u-%02u-%02u  %02u:%02u:%02u", Dow, Display.wYear, Display.wMonth, Display.wDay, Display.wHour, Display.wMinute, Display.wSecond );

	if ( How == UTC )
		Printf( L"  UTC" );

	return CL_EndLine();
}



int DisplayFileTimestamps( LPCTSTR Filename, _FileInfo_TimeDisplay How, unsigned int Which )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;

	if ( ( Which & FILEINFO_TIME_ALL ) == 0 )
		return 0;


	WIN32_FILE_ATTRIBUTE_DATA FileInfo;
	WIN32_FIND_DATA DirFileInfo;


	if ( GetFileAttributesEx( Filename, GetFileExInfoStandard, &FileInfo ) ) {

		if ( Which & FILEINFO_TIME_CREATE )
			if ( FI_ShowStamp( FileInfo.ftCreationTime,   L"    Created:   ", How ) == USER_ABORT )
				return USER_ABORT;

		if ( Which & FILEINFO_TIME_MODIFY )
			if ( FI_ShowStamp( FileInfo.ftLastWriteTime,  L"    Modified:  ", How ) == USER_ABORT )
				return USER_ABORT;

		if ( Which & FILEINFO_TIME_ACCESS )
			if ( FI_ShowStamp( FileInfo.ftLastAccessTime, L"    Accessed:  ", How ) == USER_ABORT )
				return USER_ABORT;

		return CL_EndLine();
	}
	else if ( FI_GetDirectoryInfo( Filename, &DirFileInfo ) > 0 ) {			//	GetFileAttributesEx() failed; try reading from the directory instead :

		if ( Which & FILEINFO_TIME_CREATE )
			if ( FI_ShowStamp( DirFileInfo.ftCreationTime,		L"    Created:   ", How ) == USER_ABORT )
				return USER_ABORT;

		if ( Which & FILEINFO_TIME_MODIFY )
			if ( FI_ShowStamp( DirFileInfo.ftLastWriteTime,		L"    Modified:  ", How ) == USER_ABORT )
				return USER_ABORT;

		if ( Which & FILEINFO_TIME_ACCESS )
			if ( FI_ShowStamp( DirFileInfo.ftLastAccessTime,	L"    Accessed:  ", How ) == USER_ABORT )
				return USER_ABORT;

		return CL_EndLine();
	}
	else
		return 1;
}



int DisplayFileAttributes( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	WIN32_FIND_DATA DirFileInfo;
	DWORD Attribs		= 0;
	int Count			= 0;
	
	
	Attribs = GetFileAttributes( Filename );
	if ( Attribs == INVALID_FILE_ATTRIBUTES ) {		//	GetFileAttributes() failed; try reading from the directory instead :
		if ( FI_GetDirectoryInfo( Filename, &DirFileInfo ) <= 0 )
			return 1;

		Attribs = DirFileInfo.dwFileAttributes;
	}


	QPuts( L"    Attributes:  " );

	if ( Attribs & FILE_ATTRIBUTE_READONLY ) {
		QPuts( L"Read-Only  " );
		Count++;
	}

	if ( Attribs & FILE_ATTRIBUTE_HIDDEN ) {
		QPuts( L"Hidden  " );
		Count++;
	}

	if ( Attribs & FILE_ATTRIBUTE_SYSTEM ) {
		QPuts( L"System  " );
		Count++;
	}

	if ( Attribs & FILE_ATTRIBUTE_ARCHIVE ) {
		QPuts( L"Archive  " );
		Count++;
	}

	if ( Attribs & FILE_ATTRIBUTE_NOT_CONTENT_INDEXED ) {
		QPuts( L"Index-Inhibit  " );
		Count++;
	}

	if ( Attribs & FILE_ATTRIBUTE_COMPRESSED ) {
		QPuts( L"Compressed  " );
		Count++;
	}

	if ( Attribs & FILE_ATTRIBUTE_ENCRYPTED ) {
		QPuts( L"Encrypted  " );
		Count++;
	}

	if ( !Count )
		QPuts( L"None." );

	return CL_EndLine( 2 );
}



int FI_BytesToHexString( BYTE *Bytes, unsigned int n, LPTSTR outString, size_t buflen )
{
	if ( !Bytes || !outString || !buflen )
		return -1;

	if ( ( 2 * n ) >= buflen )
		return ERROR_BUFFER_OVERFLOW;


	for ( unsigned int i = 0; i < n; i++ )
		wsprintf( outString + ( 2 * i ), L"%02X", Bytes[n - i - 1] );

	return 0;
}



int DisplayChecksums( LPCTSTR Filename, _FileInfoArgs *Args )
{
	if ( !Filename || !Args )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;

	if ( Args->Hash == 0 )
		return 0;


	wchar_t temp[SHORT_BUF_LEN]			= L"";

	DWORD crc				= 0;
	BYTE Hash[64];
	DWORD HashSize			= 0;
	int k					= 0;


	if ( !QueryIsFile( (LPTSTR) Filename ) )
		return ERROR_FILE_NOT_FOUND;


	if ( Args->Hash & FILEINFO_HASH_CRC32 ) {
		if ( !ComputeFileCRC32( Filename, &crc, NULL ) ) {
			Printf( L"    CRC-32 :  %08X", crc );
			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;

			k++;
		}
	}

	if ( Args->Hash & ( FILEINFO_HASH_CKSUM | FILEINFO_HASH_CKSUMX ) ) {
		if ( !ComputeFileCksum( Filename, &crc, NULL ) ) {
			if ( Args->Hash & FILEINFO_HASH_CKSUM ) {
				Printf( L"    cksum  :  %u", crc );
				if ( CL_EndLine() == USER_ABORT )
					return USER_ABORT;
			}

			if ( Args->Hash & FILEINFO_HASH_CKSUMX ) {
				Printf( L"    cksum  :  %08X", crc );
				if ( CL_EndLine() == USER_ABORT )
					return USER_ABORT;
			}

			k++;
		}
	}

	if ( Args->Hash & FILEINFO_HASH_MD5 ) {
		if ( !ComputeFileMD5( Filename, Hash, &HashSize ) ) {
			FI_BytesToHexString( Hash, HashSize, temp, SHORT_BUF_LEN );
			Printf( L"    MD5    :  %s", temp );
			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;

			k++;
		}
	}

	if ( Args->Hash & FILEINFO_HASH_SHA1 ) {
		if ( !ComputeFileSHA1( Filename, Hash, &HashSize ) ) {
			FI_BytesToHexString( Hash, HashSize, temp, SHORT_BUF_LEN );
			Printf( L"    SHA-1  :  %s", temp );
			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;

			k++;
		}
	}

	if ( Args->Hash & FILEINFO_HASH_SHA256 ) {
		if ( !ComputeFileSHA256( Filename, Hash, &HashSize ) ) {
			FI_BytesToHexString( Hash, HashSize, temp, SHORT_BUF_LEN );
			Printf( L"    SHA-256:  %s", temp );
			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;

			k++;
		}
	}

	if ( Args->Hash & FILEINFO_HASH_SHA384 ) {
		if ( !ComputeFileSHA384( Filename, Hash, &HashSize ) ) {
			FI_BytesToHexString( Hash, HashSize, temp, SHORT_BUF_LEN );
			Printf( L"    SHA-384:  %s", temp );
			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;

			k++;
		}
	}

	if ( Args->Hash & FILEINFO_HASH_SHA512 ) {
		if ( !ComputeFileSHA512( Filename, Hash, &HashSize ) ) {
			FI_BytesToHexString( Hash, HashSize, temp, SHORT_BUF_LEN );
			Printf( L"    SHA-512:  %s", temp );
			if ( CL_EndLine() == USER_ABORT )
				return USER_ABORT;

			k++;
		}
	}


	if ( k )
		return CL_EndLine();
	else
		return 0;
}



int Display4DosDescription( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t Description[MAX_FILENAME_LEN + 15]		= L"";


	if ( !QueryIsFile( (LPTSTR) Filename ) )
		return ERROR_FILE_NOT_FOUND;

	wsprintf( Description, L"%%@descript[\"%s\"]", Filename );
	ExpandVariables( Description, 1 );

	if ( FI_NonBlanks( Description ) ) {
		Printf( L"    4DOS file description:  %s", Description );
		return CL_EndLine( 2 );
	}

	return 0;
}



int DisplayExeInfo( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t Extension[MAX_FILENAME_LEN]		= L"";

	DWORD BinaryType			= 0;
	WORD MachineType			= 0;
	WORD Subsystem				= 0;
	int rv						= 0;


	if ( !QueryIsFile( (LPTSTR) Filename ) )
		return ERROR_FILE_NOT_FOUND;

	FH_ExtensionPart( Filename, Extension );

	if ( GetBinaryType( Filename, &BinaryType ) ) {
		switch ( BinaryType ) {

		case SCS_DOS_BINARY :
			rv = 1;
			break;

		case SCS_PIF_BINARY :
			rv = 2;
			break;

		case SCS_WOW_BINARY :
			rv = 3;
			break;

		case SCS_OS216_BINARY :
			rv = 5;
			break;

		case SCS_32BIT_BINARY :
			rv = 6;
			break;

		case SCS_64BIT_BINARY :
			rv = 9;
			break;

		case SCS_POSIX_BINARY :
			rv = 8;
			break;
		}

		if ( ( rv == 1 ) || ( rv == 5 ) || ( rv == 6 ) || ( rv == 9 ) ) {
			if ( !FI_ParseExeHeader( Filename, &MachineType, &Subsystem ) ) {

				if ( ( MachineType == IMAGE_FILE_MACHINE_I386 ) && ( Subsystem == IMAGE_VXD_SIGNATURE ) )
					rv = 4;
				else if ( ( MachineType == IMAGE_FILE_MACHINE_I386 ) && ( Subsystem == IMAGE_SUBSYSTEM_WINDOWS_GUI ) )
					rv = 6;
				else if ( ( MachineType == IMAGE_FILE_MACHINE_I386 ) && ( Subsystem == IMAGE_SUBSYSTEM_WINDOWS_CUI ) )
					rv = 7;
				else if ( ( MachineType == IMAGE_FILE_MACHINE_AMD64 ) && ( Subsystem == IMAGE_SUBSYSTEM_WINDOWS_GUI ) )
					rv = 9;
				else if ( ( MachineType == IMAGE_FILE_MACHINE_AMD64 ) && ( Subsystem == IMAGE_SUBSYSTEM_WINDOWS_CUI ) )
					rv = 10;
				else if ( Subsystem == IMAGE_SUBSYSTEM_EFI_APPLICATION )
					rv = 11;
				else if ( Subsystem == IMAGE_SUBSYSTEM_EFI_BOOT_SERVICE_DRIVER )
					rv = 12;
				else if ( Subsystem == IMAGE_SUBSYSTEM_EFI_RUNTIME_DRIVER )
					rv = 13;
				else if ( Subsystem == IMAGE_SUBSYSTEM_EFI_ROM )
					rv = 14;
				else if ( Subsystem == IMAGE_SUBSYSTEM_XBOX )
					rv = 15;
				else if ( Subsystem == IMAGE_SUBSYSTEM_WINDOWS_BOOT_APPLICATION )
					rv = 16;
			}
		}
	}
	else if ( !_wcsicmp( Extension, L".dll" ) ) {
		if ( !FI_ParseExeHeader( Filename, &MachineType, &Subsystem ) ) {
			if ( ( MachineType == IMAGE_FILE_MACHINE_I386 ) && Subsystem )
				rv = 17;
			else if ( ( MachineType == IMAGE_FILE_MACHINE_AMD64 ) && Subsystem )
				rv = 18;
		}
	}

	if ( rv && ( rv <= 18 ) ) {
		Printf( L"    Executable type ......... %s", FI_ExeTypeStr[rv] );
		return CL_EndLine( 2 );
	}

	return 0;
}



BOOL IsMatchingFileType( LPCTSTR Filename, LPCTSTR FileTypes )
{
	wchar_t Ext[SHORT_BUF_LEN]				= L"";

	LPCTSTR Match				= FileTypes;
	size_t ExtLen				= 0;


	if ( !FH_ExtensionPart( Filename, Ext ) )
		return FALSE;

	if ( ( Ext[0] != '.' ) || ( Ext[1] == '\0' ) )
		return FALSE;

	ExtLen = wcslen( Ext );


	while ( Match = wcsstr( Match, Ext ) ) {
		if ( ( Match[ExtLen] == ';' ) || ( Match[ExtLen] == '\0' ) )
			return TRUE;

		Match++;
	}

	return FALSE;
}



int DisplayTextEncoding( LPCTSTR Filename, LPCTSTR FileTypes )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;

	
	wchar_t FileEncoding[SHORT_BUF_LEN]			= L"";

	_MMFile MMFile;


	if ( !IsMatchingFileType( Filename, FileTypes ) )
		return 0;

	if ( OpenMMFile( &MMFile, Filename, MM_FLAG_ALLOW_EBCDIC ) )
		return 0;

	if ( MMFile.FileEncoding == FILE_OEM )
		if ( LooksLikeAscii( &MMFile ) )
			MMFile.FileEncoding = FILE_ASCII;

	MM_GetEncodingString( &MMFile, TRUE, FALSE, FileEncoding, SHORT_BUF_LEN );
	CloseMMFile( &MMFile );

	if ( FileEncoding[0] ) {
		Printf( L"    Text encoding ........... %s", FileEncoding );
		return CL_EndLine( 2 );
	}

	return 0;
}



int DisplayLnkTarget( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;


	wchar_t Target[MAX_FILENAME_LEN]		= L"";
	wchar_t Arguments[MAX_ARG_LEN]			= L"";

	LPTSTR Ext			= FH_ExtensionPart( Filename, NULL );
	int rv				= 0;


	if ( !Ext )
		return 0;
	else if ( _wcsicmp( Ext, L".lnk" ) )
		return 0;

	rv = GetLnkTarget( Filename, Target, Arguments );

	if ( Target[0] ) {
		if ( rv == 1 )
			Printf( L"    Link target ............. \"%s\" %s", Target, Arguments );
		else
			Printf( L"    Link target ............. %s", Target );

		return CL_EndLine( 2 );
	}

	return 0;
}



int CopySansSpaces( LPTSTR inString, LPTSTR outString, unsigned int BufLen )
{
	if ( !inString || !outString )
		return -1;


	unsigned int i = 0, j = 0;
	wchar_t ch		= '\0';

	
	do {
		ch = inString[i++];
		if ( iswspace( ch ) )
			continue;

		if ( j < ( BufLen - 1 ) ) {
			outString[j++] = ch;
			outString[j] = '\0';
		}
	}
	while ( ch );

	return j;
}



int FI_FillChar( LPTSTR String, wchar_t ch, unsigned int len )
{
	if ( !String || !ch || !len )
		return -1;


	for ( unsigned int i = 0; i < len; i++ )
		String[i] = ch;

	String[len] = '\0';

	return 0;
}



int DisplayVerInfoItem( void *Buffer, LPTSTR ItemName )
{
	if ( !Buffer || !ItemName )
		return -1;
	else if ( ItemName[0] == '\0' )
		return -1;


	wchar_t FieldName[SHORT_BUF_LEN]	= L"";
	wchar_t Value[SHORT_BUF_LEN]		= L"";
	wchar_t Dots[SHORT_BUF_LEN]			= L"";


	CopySansSpaces( ItemName, FieldName, SHORT_BUF_LEN );

	FI_GetFileVerString( Buffer, FieldName, Value, SHORT_BUF_LEN );

	if ( !FI_NonBlanks( Value ) )
		return 1;

	FI_FillChar( Dots, '.', 24 - (unsigned int) wcslen( ItemName ) );
	Printf( L"    %s %s %s", ItemName, Dots, Value );

	return CL_EndLine();
}



int DisplayVerInfo( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	DWORD VerInfoSize					= 0;
	void *Buffer						= NULL;
	VS_FIXEDFILEINFO *FileInfo			= NULL;
	UINT FileInfoSize					= 0;
	int Count							= 0;
	int t								= 0;


	if ( !QueryIsFile( (LPTSTR) Filename ) )
		return ERROR_FILE_NOT_FOUND;


	VerInfoSize = GetFileVersionInfoSize( Filename, NULL );
	if ( !VerInfoSize )
		return ERROR_MOD_NOT_FOUND;

	Buffer = malloc( VerInfoSize );
	if ( !Buffer )
		return ERROR_NOT_ENOUGH_MEMORY;


	if ( GetFileVersionInfo( Filename, 0, VerInfoSize, Buffer ) ) {

		VerQueryValue( Buffer, L"\\", (void**) &FileInfo, &FileInfoSize );

		if ( ( t = DisplayVerInfoItem( Buffer, L"Product Name" ) ) == 0 )
			Count++;
		else if ( t == USER_ABORT )
			goto UserAbort;

		if ( ( t = DisplayVerInfoItem( Buffer, L"Internal Name" ) ) == 0 )
			Count++;
		else if ( t == USER_ABORT )
			goto UserAbort;

		if ( ( t = DisplayVerInfoItem( Buffer, L"Original Filename" ) ) == 0 )
			Count++;
		else if ( t == USER_ABORT )
			goto UserAbort;

		if ( FileInfo ) {
			unsigned int Version[4]			= { 0, 0, 0, 0 };

			Version[0]		= FileInfo->dwFileVersionMS >> 16;
			Version[1]		= FileInfo->dwFileVersionMS & 0xffff;
			Version[2]		= FileInfo->dwFileVersionLS >> 16;
			Version[3]		= FileInfo->dwFileVersionLS & 0xffff;

			Printf( L"    Version ................. %u.%u.%u.%u", Version[0], Version[1], Version[2], Version[3] );
			if ( CL_EndLine() == USER_ABORT )
				goto  UserAbort;

			Count++;
		}

		if ( ( t = DisplayVerInfoItem( Buffer, L"Product Version" ) ) == 0 )
			Count++;
		else if ( t == USER_ABORT )
			goto UserAbort;

		if ( ( t = DisplayVerInfoItem( Buffer, L"File Version" ) ) == 0 )
			Count++;
		else if ( t == USER_ABORT )
			goto UserAbort;

		if ( ( t = DisplayVerInfoItem( Buffer, L"Build" ) ) == 0 )
			Count++;
		else if ( t == USER_ABORT )
			goto UserAbort;

		if ( ( t = DisplayVerInfoItem( Buffer, L"Company Name" ) ) == 0 )
			Count++;
		else if ( t == USER_ABORT )
			goto UserAbort;

		if ( ( t = DisplayVerInfoItem( Buffer, L"Language" ) ) == 0 )
			Count++;
		else if ( t == USER_ABORT )
			goto UserAbort;

		if ( ( t = DisplayVerInfoItem( Buffer, L"Legal Trademarks" ) ) == 0 )
			Count++;
		else if ( t == USER_ABORT )
			goto UserAbort;
	}

	if ( Count )
		if ( CL_EndLine() == USER_ABORT )
			goto UserAbort;

	free( Buffer );
	return 0;


UserAbort:

	free( Buffer );
	return USER_ABORT;
}



int DisplayNameAndValue( LPTSTR Name, LPTSTR Value )
{
	if ( !Name || !Value )
		return -1;


	wchar_t Dots[SHORT_BUF_LEN]			= L"";


	FI_FillChar( Dots, '.', 24 - (unsigned int) wcslen( Name ) );
	Printf( L"    %s %s %s", Name, Dots, Value );
	return CL_EndLine();
}



#ifdef IMGINFO_VER


BOOL FI_Leapness( int year )
{
	if ( ( year % 400 ) == 0 ) return TRUE;
	if ( ( year % 100 ) == 0 ) return FALSE;
	return ( ( year % 4 ) == 0 );
}



long int FI_Julian( int year, int month, int day )
{
	if ( ( year < EPOCH ) || ( year > ULTIMA ) )
		return -1;

	if ( ( month < 1 ) || ( month > 12 ) )
		return -1;

	BOOL leap = FI_Leapness( year );

	if ( ( day < 1 ) || ( day > FI_MonthLength[month][leap] ) )
		return -1;

	long int j = ( ( year - EPOCH ) * 365 );
	j += ( ( year - EPOCH ) / 4 );
	j -= ( ( year - EPOCH ) / 100 );
	j += ( ( year - EPOCH ) / 400 );
	j += FI_Kalends[month][leap];
	j += day;

	return j;
}



int FI_ExifTimeToSystemTime( LPCTSTR ExifStamp, SYSTEMTIME *SystemTime )
{
	if ( !ExifStamp || !SystemTime )
		return 0;
	else if ( ExifStamp[0] == '\0' )
		return 0;


	unsigned int Year			= 0;
	unsigned int Month			= 0;
	unsigned int Date			= 0;
	unsigned int Hour			= 0;
	unsigned int Minute			= 0;
	unsigned int Second			= 0;
	int Julian					= 0;

	wchar_t *ch					= NULL;


	_set_errno( 0 );

	Year	= wcstoul( ExifStamp, &ch, 10 );
	if ( ( Year < EPOCH ) || ( Year > ULTIMA ) || errno )
		return 0;
	else if ( !iswpunct( *ch ) )
		return 0;

	Month	= wcstoul( ch + 1, &ch, 10 );
	if ( ( Month < 1 ) || ( Month > 12 ) || errno )
		return 0;
	else if ( !iswpunct( *ch ) )
		return 0;

	Date	= wcstoul( ch + 1, &ch, 10 );
	if ( ( Date < 1 ) || ( Date > 31 ) || errno )
		return 0;
	else if ( !iswspace( *ch ) )
		return 0;

	while ( iswspace( *ch ) )
		ch++;

	Hour	= wcstoul( ch, &ch, 10 );
	if ( ( Hour > 23 ) || errno )
		return 0;
	else if ( !iswpunct( *ch ) )
		return 0;

	Minute	= wcstoul( ch + 1, &ch, 10 );
	if ( ( Minute > 59 ) || errno )
		return 0;
	else if ( !iswpunct( *ch ) )
		return 0;

	Second	= wcstoul( ch + 1, &ch, 10 );
	if ( ( Second > 59 ) || errno )
		return 0;
	

	Julian = FI_Julian( Year, Month, Date );
	if ( Julian < 0 )
		return 0;


	memset( SystemTime, 0x00, sizeof( SYSTEMTIME ) );

	SystemTime->wYear		= Year;
	SystemTime->wMonth		= Month;
	SystemTime->wDay		= Date;
	SystemTime->wHour		= Hour;
	SystemTime->wMinute		= Minute;
	SystemTime->wSecond		= Second;
	SystemTime->wDayOfWeek	= Julian % 7;
	return 1;
}



int DisplayImageFileInfo( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t temp[MAX_ARG_LEN]	= L"";

	Image *SourceImage			= NULL;

	UINT HSize			= 0;
	UINT VSize			= 0;
	PixelFormat Fmt		= 0;

	int rv				= 0;


	OpenImageFile( Filename, &SourceImage, FALSE );

	if ( SourceImage ) {
		HSize = SourceImage->GetWidth();
		VSize = SourceImage->GetHeight();
		Fmt = SourceImage->GetPixelFormat();

		if ( HSize && VSize ) {
			/*
			Printf( L"    Image file properties:" );
			if ( ( rv = CL_EndLine() ) == USER_ABORT )
				goto Quit;
			*/

			if ( Fmt & 0xff00 )
				wsprintf( temp, L"%u \u00d7 %u    %u BPP", HSize, VSize, ( Fmt & 0xff00 ) >> 8 );
			else
				wsprintf( temp, L"%u \u00d7 %u", HSize, VSize );

			if ( ( rv = DisplayNameAndValue( L"Image size", temp ) ) == USER_ABORT )
				goto Quit;


			temp[0] = '\0';
			GetExifString( SourceImage, PropertyTagExifDTOrig, temp );

			if ( LooksLikeExifDateTimeString( temp ) ) {
				SYSTEMTIME SysTime;

				if ( FI_ExifTimeToSystemTime( temp, &SysTime ) > 0 )
					wsprintf( temp, L"%s %04u-%02u-%02u  %02u:%02u:%02u", FI_DowNames[SysTime.wDayOfWeek], SysTime.wYear, SysTime.wMonth, SysTime.wDay, SysTime.wHour, SysTime.wMinute, SysTime.wSecond );
				else {
					temp[4] = '-';
					temp[7] = '-';
				}

				if ( ( rv = DisplayNameAndValue( L"Date taken", temp ) ) == USER_ABORT )
					goto Quit;
			}

			temp[0] = '\0';
			GetExifString( SourceImage, PropertyTagArtist, temp );

			if ( FI_NonBlanks( temp ) && wcscmp( temp, II_NotAvailable ) ) {
				if ( ( rv = DisplayNameAndValue( L"Artist", temp ) ) == USER_ABORT )
					goto Quit;
			}

			temp[0] = '\0';
			GetExifString( SourceImage, PropertyTagCopyright, temp );

			if ( FI_NonBlanks( temp ) && wcscmp( temp, II_NotAvailable ) ) {
				if ( ( rv = DisplayNameAndValue( L"Copyright", temp) ) == USER_ABORT )
					goto Quit;
			}

			temp[0] = '\0';
			GetExifString( SourceImage, PropertyTagSoftwareUsed, temp );

			if ( FI_NonBlanks( temp ) && wcscmp( temp, II_NotAvailable ) ) {
				if ( ( rv = DisplayNameAndValue( L"Software", temp ) ) == USER_ABORT )
					goto Quit;
			}


			if ( ( rv = CL_EndLine() ) == USER_ABORT )
				goto Quit;

			rv = 0;
		}
	}


Quit:

	if ( SourceImage )
		delete SourceImage;

	return rv;
}


#endif	//	IMGINFO_VER



int FI_GetPropString( IPropertyStore *Store, REFPROPERTYKEY Key, LPTSTR String, UINT len )
{
	if ( !Store || !String || !len )
		return -1;


	PROPVARIANT Var;


	memset( &Var, 0x00, sizeof( PROPVARIANT ) );
	String[0] = '\0';

	if ( !SUCCEEDED( Store->GetValue( Key, &Var ) ) )
		return 1;

	if ( PropVariantToString( Var, String, len ) ) {
		String[0] = '\0';
		return 2;
	}

	return 0;
}



int FI_GetPropUnsigned( IPropertyStore *Store, REFPROPERTYKEY Key, unsigned int *outInt )
{
	if ( !Store || !outInt )
		return -1;


	PROPVARIANT Var;


	memset( &Var, 0x00, sizeof( PROPVARIANT ) );
	*outInt = 0;

	if ( !SUCCEEDED( Store->GetValue( Key, &Var ) ) )
		return 1;

	if ( Var.vt != VT_UI4 )
		return 2;

	*outInt = Var.uintVal;
	return 0;
}



int DisplayWinProps( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	wchar_t temp[SHORT_BUF_LEN]		= L"";
	IPropertyStore *Store			= NULL;
	PROPVARIANT Var;

	unsigned int Count		= 0;
	unsigned int UInt		= 0;
	int rv					= 0;


	SHGetPropertyStoreFromParsingName( Filename, NULL, GPS_DEFAULT, __uuidof( IPropertyStore ), (void**) &Store );

	if ( Store ) {

		FI_GetPropString( Store, PKEY_Title, temp, SHORT_BUF_LEN );
		if ( FI_NonBlanks( temp ) ) {
			if ( ( rv = DisplayNameAndValue( L"Title", temp ) ) == USER_ABORT )
				goto Quit;

			Count++;
		}

		FI_GetPropString( Store, PKEY_Media_SubTitle, temp, SHORT_BUF_LEN );
		if ( FI_NonBlanks( temp ) ) {
			if ( ( rv = DisplayNameAndValue( L"Subtitle", temp ) ) == USER_ABORT )
				goto Quit;

			Count++;
		}

		FI_GetPropString( Store, PKEY_Comment, temp, SHORT_BUF_LEN );
		if ( FI_NonBlanks( temp ) ) {
			if ( ( rv = DisplayNameAndValue( L"Comment", temp ) ) == USER_ABORT )
				goto Quit;

			Count++;
		}

		FI_GetPropString( Store, PKEY_Music_AlbumArtist, temp, SHORT_BUF_LEN );
		if ( FI_NonBlanks( temp ) ) {
			if ( ( rv = DisplayNameAndValue( L"Album artist", temp ) ) == USER_ABORT )
				goto Quit;

			Count++;
		}

		FI_GetPropString( Store, PKEY_Music_AlbumTitle, temp, SHORT_BUF_LEN );
		if ( FI_NonBlanks( temp ) ) {
			if ( ( rv = DisplayNameAndValue( L"Album title", temp ) ) == USER_ABORT )
				goto Quit;

			Count++;
		}

		FI_GetPropString( Store, PKEY_Music_Artist, temp, SHORT_BUF_LEN );
		if ( FI_NonBlanks( temp ) ) {
			if ( ( rv = DisplayNameAndValue( L"Artist", temp ) ) == USER_ABORT )
				goto Quit;

			Count++;
		}

		FI_GetPropString( Store, PKEY_Music_Genre, temp, SHORT_BUF_LEN );
		if ( FI_NonBlanks( temp ) ) {
			if ( ( rv = DisplayNameAndValue( L"Genre", temp ) ) == USER_ABORT )
				goto Quit;

			Count++;
		}

		memset( &Var, 0x00, sizeof( PROPVARIANT ) );
		if ( SUCCEEDED( Store->GetValue( PKEY_Media_Duration, &Var ) ) ) {
			if ( Var.vt == VT_UI8 ) {
				unsigned __int64 msec		= ( Var.uhVal.QuadPart % 10000000 ) / 10000;
				unsigned __int64 sec		= Var.uhVal.QuadPart / 10000000;
				unsigned __int64 min		= 0;
				unsigned __int64 hrs		= 0;


				if ( sec >= 60 )
					min = sec / 60, sec %= 60;

				if ( min >= 60 )
					hrs = min / 60, min %= 60;

				wsprintf( temp, L"%I64u:%02I64u:%02I64u", hrs, min, sec );
				if ( ( rv = DisplayNameAndValue( L"Duration", temp ) ) == USER_ABORT )
					goto Quit;

				Count++;
			}
		}

		if ( !FI_GetPropUnsigned( Store, PKEY_Music_TrackNumber, &UInt ) ) {
			_ultow_s( UInt, temp, SHORT_BUF_LEN, 10 );
			if ( ( rv = DisplayNameAndValue( L"Track number", temp ) ) == USER_ABORT )
				goto Quit;

			Count++;
		}


		if ( Count )
			if ( ( rv = CL_EndLine() ) == USER_ABORT )
				goto Quit;
	}


Quit:

	if ( Store )
		Store->Release();

	return rv;
}



int DisplaySummaryItem( IPropertyStorage *Property, unsigned int FieldID, LPTSTR Name )
{
	if ( !Property || !Name )
		return -1;

	if ( CL_AbortFlag )
		return USER_ABORT;


	wchar_t ValueString[SHORT_BUF_LEN]		= L"";
	wchar_t Dots[SHORT_BUF_LEN]				= L"";
	PROPSPEC PropertySpec;
	PROPVARIANT PropertyReturns;
	FILETIME LocalFileTime;
	SYSTEMTIME SystemTime;

	HRESULT hr		= 0;



	PropertySpec.ulKind		= PRSPEC_PROPID;
	PropertySpec.propid		= FieldID;

	memset( &PropertyReturns, 0x00, sizeof( PROPVARIANT ) );
	
	hr = Property->ReadMultiple( 1, &PropertySpec, &PropertyReturns );
	if ( FAILED( hr ) )
		return 1;

	
	switch ( PropertyReturns.vt ) {

	case VT_LPSTR :
		MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, PropertyReturns.pszVal, -1, ValueString, SHORT_BUF_LEN );
		break;

	case VT_LPWSTR :
		wcscpy_s( ValueString, SHORT_BUF_LEN, PropertyReturns.pwszVal );
		break;

	case VT_I2 :
		wsprintf( ValueString, L"%d", PropertyReturns.iVal );
		break;

	case VT_UI2 :
		wsprintf( ValueString, L"%u", PropertyReturns.uiVal );
		break;

	case VT_I4 :
		wsprintf( ValueString, L"%d", PropertyReturns.lVal );
		break;

	case VT_UI4 :
		wsprintf( ValueString, L"%u", PropertyReturns.ulVal );
		break;

	case VT_FILETIME :

		if ( FieldID == 10 ) {
			//		'Edit Time' is an absolute length, not a moment to be translated:

			unsigned __int64 mSec = 0, Seconds = 0, Minutes = 0, Hours = 0;

			mSec = ( (unsigned __int64) PropertyReturns.filetime.dwHighDateTime << 32 ) | PropertyReturns.filetime.dwLowDateTime;
			mSec /= 10000;

			Seconds = mSec / 1000;
			mSec %= 1000;

			Minutes = Seconds / 60;
			Seconds %= 60;

			Hours = Minutes / 60;
			Minutes %= 60;

			wsprintf( ValueString, L"%02I64u:%02I64u:%02I64u", Hours, Minutes, Seconds );
		}
		else {
			//		Other times are UTC times, so translate to local time:

			FileTimeToLocalFileTime( &PropertyReturns.filetime, &LocalFileTime );
			FileTimeToSystemTime( &LocalFileTime, &SystemTime );
			wsprintf( ValueString, L"%04u-%02u-%02u %02u:%02u.%02u", SystemTime.wYear, SystemTime.wMonth, SystemTime.wDay, SystemTime.wHour, SystemTime.wMinute, SystemTime.wSecond );
		}


		break;

	default :
		ValueString[0] = '\0';
	}


	if ( FI_NonBlanks( ValueString ) == '\0' )
		return 2;

	FI_FillChar( Dots, '.', (unsigned int) ( 24 - wcslen( Name ) ) );
	Printf( L"    %s %s %s", Name, Dots, ValueString );
	return CL_EndLine();
}



int DisplaySummaryInfo( LPCTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	IPropertySetStorage *PropertySet		= NULL;
	IPropertyStorage *Property				= NULL;

	int Count		= 0;



	if ( StgOpenStorageEx( Filename, STGM_READ | STGM_SHARE_EXCLUSIVE, STGFMT_ANY, 0, NULL, 0, IID_IPropertySetStorage, (void**) &PropertySet ) == S_OK ) {

		if ( PropertySet->Open( FMTID_SummaryInformation, STGM_READ | STGM_SHARE_EXCLUSIVE, &Property ) == S_OK ) {

			if ( DisplaySummaryItem( Property, 2, L"Title" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 3, L"Subject" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 4, L"Author" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 5, L"Keywords" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 6, L"Comments" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 7, L"Template" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 8, L"Last Author" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 9, L"Revision Number" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 10, L"Edit Time" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 11, L"Last Printed" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 12, L"Created" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 13, L"Last Saved" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 14, L"Page Count" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 15, L"Word Count" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 16, L"Char Count" ) == 0 )
				Count++;

			if ( DisplaySummaryItem( Property, 18, L"App Name" ) == 0 )
				Count++;

			Property->Release();
		}

		PropertySet->Release();
	}

	if ( CL_AbortFlag )
		return USER_ABORT;

	if ( Count )
		return CL_EndLine();

	return 0;
}



int ShowFileInfo( LPCTSTR Filename, int Flags, void *MoreArgs )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return -1;


	if ( !QueryIsFile( (LPTSTR) Filename ) )
		return ERROR_FILE_NOT_FOUND;


	_FileInfoArgs *Args		= (_FileInfoArgs*) MoreArgs;
	LPTSTR Ext				= FH_ExtensionPart( Filename, NULL );
	

	if ( HLputs( (LPTSTR) Filename, 2 ) == USER_ABORT )
		return USER_ABORT;


	try {

		if ( !( Flags & FILEINFO_OMIT_VERHEADER ) )
			if ( DisplayFileVerHeader( Filename ) == USER_ABORT )
				return USER_ABORT;

		if ( !( Flags & FILEINFO_OMIT_SIZE ) )
			if ( DisplayFileSizeCosmetic( Filename ) == USER_ABORT )
				return USER_ABORT;

		if ( !( Flags & FILEINFO_OMIT_TIMESTAMPS ) )
			if ( DisplayFileTimestamps( Filename, Args->TimeDisplay, Args->TimeStamps ) == USER_ABORT )
				return USER_ABORT;

		if ( !( Flags & FILEINFO_OMIT_ATTRIBUTES ) )
			if ( DisplayFileAttributes( Filename ) == USER_ABORT )
				return USER_ABORT;

		if ( !( Flags & FILEINFO_OMIT_CHECKSUMS ) )
			if ( DisplayChecksums( Filename, Args ) == USER_ABORT )
				return USER_ABORT;

		if ( !( Flags & FILEINFO_OMIT_4DESCRIPTION ) )
			if ( Display4DosDescription( Filename ) == USER_ABORT )
				return USER_ABORT;

		if ( !( Flags & FILEINFO_OMIT_EXE_INFO ) )
			if ( DisplayExeInfo( Filename ) == USER_ABORT )
				return USER_ABORT;

		if ( !( Flags & FILEINFO_OMIT_TEXT_ENCODING ) )
			if ( DisplayTextEncoding( Filename, Args->TextFileTypes ) == USER_ABORT )
				return USER_ABORT;

		if ( !( Flags & FILEINFO_OMIT_LNK_TARGET ) && Ext )
			if ( !_wcsicmp( Ext, L".lnk" ) )
				if ( DisplayLnkTarget( Filename ) == USER_ABORT )
					return USER_ABORT;

		if ( !( Flags & FILEINFO_OMIT_VERINFO ) )
			if ( DisplayVerInfo( Filename ) == USER_ABORT )
				return USER_ABORT;

#ifdef IMGINFO_VER

		if ( !( Flags & FILEINFO_OMIT_IMAGE_INFO ) )
			if ( ImageFileType( Filename ) )
				if ( DisplayImageFileInfo( Filename ) == USER_ABORT )
					return USER_ABORT;

#endif	//	IMGINFO_VER

		if ( !( Flags & FILEINFO_OMIT_WIN_PROPS ) )
			if ( DisplayWinProps( Filename ) == USER_ABORT )
				return USER_ABORT;

		if ( !(Flags & FILEINFO_OMIT_SUMMARY_INFO ) )
			if ( DisplaySummaryInfo( Filename ) == USER_ABORT )
				return USER_ABORT;

		Args->Count += 1;

	}
	catch ( int n ) {
		Printf( L"\n   *  Exception %d !\n\n", n );
		return USER_ABORT;
	}

	return 0;
}



DLLExports int WINAPI fileinfo( LPTSTR inString )
{
	if ( !inString )
		return 1;



	int Flags					= 0;
	int Recurse					= FH_FLAG_INCLUDE_LISTS | FH_FLAG_ALLOW_FILE_LISTS;
	unsigned int MinDepth		= 0;
	unsigned int MaxDepth		= UINT_MAX;
	DWORD AttribMask			= FILE_ATTRIBUTE_DIRECTORY;
	DWORD AttribValues			= 0;
	void *FirstRange			= NULL;

	_FileInfoArgs Args			= { 0, Default, 0, 0, NULL };

	wchar_t Wildspec[MAX_FILENAME_LEN]		= L"";
	wchar_t TextFileTypes[MAX_ARG_LEN]		= L"";
	
	BOOL NoColor				= FALSE;
	BOOL ShowFinalCount			= TRUE;
	unsigned int PauseLines		= 0;

	int NumFilespecs			= 0;
	int rv						= 1;
	int sfirv					= 0;
	int t						= 0;



	CL_NoPausing();

	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_EXCL_RANGE_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {
			if ( !wcsncmp( arg[i], L"/?", 2 ) ) {
				ShowCmdHelp( FileInfoHelpText, ( arg[i][2] == '?' ), __FILE__, MACRO_STR( FILEINFO_VER ) );
				rv = 0;
				goto CleanupAndExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) && wcschr( arg[i], '?' ) ) {
				QPuts( (LPTSTR) FileInfoOptNHelpText );
				rv = 0;
				goto CleanupAndExit;
			}
			else if ( t = ParseCommonOption( arg[i], FALSE, &AttribMask, &AttribValues, &Recurse, &MinDepth, &MaxDepth, &NoColor, &FirstRange ) ) {
				if ( t < 0 )
					goto CleanupAndExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				if ( wcschr( arg[i], '?' ) ) {
					QPuts( (LPTSTR) FileInfoOptCHelpText );
					rv = 0;
					goto CleanupAndExit;
				}


				CharUpperW( arg[i] );
				wchar_t *ch			= arg[i] + 2;
				size_t len			= 0;
				unsigned int v		= 0;
				if ( IsColonOrEquals( *ch ) )
					ch++;


				if ( *ch == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing hash scheme: \"%s\"\r\n", arg[i] );
					goto CleanupAndExit;
				}

				while ( *ch ) {
					while ( ( *ch == ',' ) || ( *ch == ';' ) )
						ch++;

					if ( *ch ) {
						v = 0;

						for ( int j = 0; j < FI_NumHashes; j++ ) {
							len = wcslen( FI_HashName[j] );

							if ( !wcsncmp( ch, FI_HashName[j], len ) && ( ( ch[len] == '\0' ) || ( ch[len] == ',' ) || ( ch[len] == ';' ) ) ) {
								v = FI_HashVal[j];
								break;
							}
						}

						if ( v ) {
							Args.Hash |= v;
							ch += len;
						}
						else {
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Unknown hash scheme: \"/C:%s\"\r\n", ch );
							goto CleanupAndExit;
						}
					}
				}
			}
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				NoColor = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/O", 2 ) ) {

				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				while ( arg[i][j] ) {
					switch ( towupper( arg[i][j] ) ) {

					case '?' :
						QPuts( (LPTSTR) FileInfoOptOHelpText );
						rv = 0;
						goto CleanupAndExit;

					case '@' :
						Recurse &= ~FH_FLAG_ALLOW_FILE_LISTS;
						break;

					case 'A' :
						Flags |= FILEINFO_OMIT_ATTRIBUTES;
						break;

					case 'C' :
						Flags |= FILEINFO_OMIT_CHECKSUMS;
						break;

					case 'D' :
						Flags |= FILEINFO_OMIT_4DESCRIPTION;
						break;

					case 'E' :
						Flags |= FILEINFO_OMIT_TEXT_ENCODING;
						break;

					case 'F' :
						ShowFinalCount = FALSE;
						break;

					case 'H' :
						Flags |= FILEINFO_OMIT_VERHEADER;
						break;

#ifdef IMGINFO_VER

					case 'I' :
						Flags |= FILEINFO_OMIT_IMAGE_INFO;
						break;

#endif	//	IMGINFO_VER

					case 'L' :
						Flags |= FILEINFO_OMIT_LNK_TARGET;
						break;

					case 'P' :
						Flags |= FILEINFO_OMIT_WIN_PROPS;
						break;

					case 'S' :
						Flags |= FILEINFO_OMIT_SUMMARY_INFO;
						break;

					case 'T' :
						Flags |= FILEINFO_OMIT_TIMESTAMPS;
						break;

					case 'V' :
						Flags |= FILEINFO_OMIT_VERINFO;
						break;

					case 'X' :
						Flags |= FILEINFO_OMIT_EXE_INFO;
						break;

					case 'Z' :
						Flags |= FILEINFO_OMIT_SIZE;
						break;

					default :
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Unknown suboption: /O:%c\r\n", towupper( arg[i][j] ) );
						goto CleanupAndExit;
					}
					
					j++;
				}
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				if ( CL_ParsePOption( arg[i], &PauseLines ) )
					goto CleanupAndExit;
			}
			else if ( !_wcsnicmp( arg[i], L"/T", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				while ( arg[i][j] ) {
					switch ( towupper( arg[i][j] ) ) {

					case '?' :
						QPuts( (LPTSTR) FileInfoOptTHelpText );
						goto CleanupAndExit;

					case 'A' :
						Args.TimeStamps |= FILEINFO_TIME_ACCESS;
						break;

					case 'C' :
						Args.TimeStamps |= FILEINFO_TIME_CREATE;
						break;

					case 'D' :
						Args.TimeDisplay = Default;
						break;

					case 'L' :
						Args.TimeDisplay = Smart;
						break;

					case 'M' :
						Args.TimeStamps |= FILEINFO_TIME_MODIFY;
						break;

					case 'U' :
						Args.TimeDisplay = UTC;
						break;

					default :
						CharUpperW( arg[i] );
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Unknown time mode: %s\r\n", arg[i] );
						goto CleanupAndExit;
					}

					j++;
				}
			}
			else if ( arg[i][0] == '/' ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				goto CleanupAndExit;
			}
			else
				NumFilespecs++;
		}
	}


	if ( PauseLines )
		CL_SetPauseLines( PauseLines );

	CL_PickHighlightColor( NoColor );
	CL_HookBreak();


	GetEnvironmentVariable( FI_TextFileTypesVar, TextFileTypes, MAX_ARG_LEN );


	rv = 0;


	Args.Count			= 0;
	Args.TextFileTypes	= TextFileTypes[0] ? TextFileTypes : FI_DefaultTextFileTypes;

	if ( Args.TimeStamps == 0 )
		Args.TimeStamps = FILEINFO_TIME_ALL;

	if ( Args.Hash == 0 )
		Args.Hash = FILEINFO_HASH_CRC32 | FILEINFO_HASH_MD5;



	if ( NumFilespecs ) {
		for ( int i = 0; i < numargs; i++ ) {
			if ( arglen[i] ) {
				if ( arg[i][0] != '/' ) {

					if ( ( arg[i][0] == '@' ) && arg[i][1] && ( Recurse & FH_FLAG_ALLOW_FILE_LISTS ) )
						wcscpy_s( Wildspec, MAX_FILENAME_LEN, arg[i] );
					else {
						FH_Truename( arg[i], Wildspec, MAX_FILENAME_LEN, TRUENAME_FLAG_HOME_PATHS | TRUENAME_FLAG_INCLUDE_LISTS );
						if ( QueryIsDirectory( Wildspec ) )
							AppendFilename( Wildspec, L"*", MAX_FILENAME_LEN );
					}

					sfirv = HandleMatchingFiles( Wildspec, Flags, &Args, ShowFileInfo, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );

					if ( sfirv == USER_ABORT ) {
						rv = 3;
						goto CleanupAndExit;
					}
					else if ( sfirv )
						rv = 2;
				}
			}
		}
	}
	else {
		
		sfirv = HandleMatchingFiles( L"*", Flags, &Args, ShowFileInfo, Recurse, AttribMask, AttribValues, FirstRange, MinDepth, MaxDepth );

		if ( sfirv == USER_ABORT ) {
			rv = 3;
			goto CleanupAndExit;
		}
		else if ( sfirv )
			rv = 2;
	}


	if ( ShowFinalCount ) {
		if ( Args.Count ) {
			if ( Args.Count == 1 )
				wcscpy_s( Wildspec, MAX_FILENAME_LEN, L"1 matching file found." );
			else
				wsprintf( Wildspec, L"%u matching files found.", Args.Count );

			HLputs( Wildspec, 1 );
		}
	}


CleanupAndExit:

	FH_FreeLinkedList( &FirstRange );
	CL_ReleaseBreak();
	return rv;
}


#endif	//	FILEINFO_VER


/*

	1.0.0.0		2021-09-14

	1.0.1.0		2022-03-01
				Rewritten to use newer versions of FileHandler.cpp and conlist.cpp.  Added /D and /NC to
				disable highlight.  Help text is now displayed via ShowCmdHelp().

	1.0.2.0		2022-03-02
				Added image file info; /O:I to disable.

	1.0.3.0		2022-03-03
				Added support for some shell properties; /O:P to disable.

	1.0.3.1		2022-03-04
				Now I'm calling them "Windows properties" instead of "shell properties".  No changes to
				functionality.

	1.0.3.2		2022-03-10
				Renamed the "not available" string.

	1.0.4.0		2022-03-17
				Added support for System.Music.TrackNumber.

	1.0.4.1		2023-08-01
				Fixed a possible memory leak on user abort.

	1.0.4.2		2023-08-22
				Defined filename args as constant types.

	1.0.5.0		2023-09-21
				Added executable type info, and /O:X to disable it.  Added /O? to show help text for
				/O suboptions.

	1.0.5.1		2023-10-10
				Tweaked the help text to mention the existence of /O?

	1.0.5.2		2023-12-14
				Fix for DisplaySummaryInfo() not correctly returning USER_ABORT when the user
				did in fact abort.  Renamed this file to FileInfoCmd.cpp.

	1.0.6.0		2023-12-14
				Explicitly #include "ImgInfo.cpp".
				Summary item names get a field length of 24, the same as version info, image file
				info, and Windows property names.  Added "32-bit DLL file" and "64-bit DLL file"
				executable type strings.  Added a bit more sanity checking to FI_ParseExeHeader().

	1.0.6.1		2023-12-15
				DisplayFileTimeStamps() and DisplayFileAttributes() :  If the usual methods of
				getting info fail, try reading from the directory instead.

	1.0.7.0		2023-12-15
				Added /T: to set the time stamp display mode.  Added ability to show time stamps
				as UTC.  Fixed a possible memory leak with /O?

	1.0.7.1		2023-12-16
				Removed the #include "ImgInfo.cpp"
				Fixed this file so it can be successfully compiled either with or without
				ImgInfo.cpp.

	1.0.7.2		2023-12-17
				If the EXIF date stamp looks valid, display it in the same format as other date
				stamps in this command, including day of the week.

	1.0.8.0		2024-01-03
				Changes for conlist v1.1, to better support Ctrl-C and Ctrl-Break.
				Check for display strings containing only whitespace, and don't display them.
				(Why would anyone do this?)

	1.0.8.1		2024-01-10
				Added /N? to show help text for /N suboptions.

	1.0.8.2		2024-01-17
				DLL files may have any subsystem value, so long as it isn't zero.

	1.0.9.0		2024-01-29
				Adds text file encoding, .LNK file targets.
				Enabled @file lists.

	1.0.9.1		2024-01-31
				Added /T:ACM to select which time stamps to display.

	1.0.9.2		2024-02-05
				Added /C: to select which checksums/hashes to compute and display.

	1.0.9.3		2024-02-07
				Reworked and simplified the code which displays file hashes.

*/
