
#ifndef IMGINFO_VER
#define IMGINFO_VER				1.0.0.3



const int NumGdiPlusStatusStrings		= 21;

const wchar_t GdiPlusStatusStrings[NumGdiPlusStatusStrings][26] = { L"Ok", L"GenericError", L"InvalidParameter", L"OutOfMemory", L"ObjectBusy",
		L"InsufficientBuffer", L"NotImplemented", L"Win32Error", L"WrongState", L"Aborted", L"FileNotFound", L"ValueOverflow", L"AccessDenied",
		L"UnknownImageFormat", L"FontFamilyNotFound", L"NotTrueTypeFont", L"UnsupportedGdiplusVersion", L"GdiplusNotInitialized",
		L"PropertyNotFound", L"PropertyNotSupported", L"ProfileNotFound" };

const wchar_t II_NotAvailable[]		= L"N/A";



int ImageFileType( LPCTSTR Filename )
{
	if ( !Filename )
		return 0;

	LPTSTR Ext = FH_ExtensionPart( Filename, NULL );

	if ( !Ext )
		return 0;

	if ( !_wcsicmp( Ext, L".bmp" ) )
		return 1;

	if ( !_wcsicmp( Ext, L".jpg" ) || !_wcsicmp( Ext, L".jpeg" ) )
		return 2;

	if ( !_wcsicmp( Ext, L".png" ) )
		return 3;

	if ( !_wcsicmp( Ext, L".gif" ) )
		return 4;

	if ( !_wcsicmp( Ext, L".tif" ) || !_wcsicmp( Ext, L".tiff" ) )
		return 5;

	if ( !_wcsicmp( Ext, L".ico" ) )
		return 6;

	if ( !_wcsicmp( Ext, L".emf" ) )
		return 7;

	if ( !_wcsicmp( Ext, L".wmf" ) )
		return 8;

	return 0;
}



int GetExifShort( Image *SourceImage, PROPID WhichTag )
{
	if ( !SourceImage )
		return -1;

	UINT PropertySize		= 0;
	PropertyItem *buffer	= NULL;
	int rv					= -1;

	PropertySize = SourceImage->GetPropertyItemSize( WhichTag );
	if ( PropertySize == 0 )
		return -1;

	buffer = (PropertyItem*) malloc( PropertySize );
	if ( !buffer )
		return -1;

	if ( SourceImage->GetPropertyItem( WhichTag, PropertySize, buffer ) ) {
		free( buffer );
		return -1;
	}

	switch ( buffer->type ) {
		case PropertyTagTypeByte :
			rv = *( (BYTE*) buffer->value );
			break;

		case PropertyTagTypeShort :
			rv = *( (WORD*) buffer->value );
			break;

		case PropertyTagTypeLong :
			rv = *( (DWORD*) buffer->value );
			break;
	}

	free( buffer );
	return rv;
}



int GetExifString( Image *SourceImage, PROPID WhichTag, LPTSTR outString )
{
	if ( !SourceImage || !outString )
		return -1;


	UINT PropertySize		= 0;
	PropertyItem *buffer	= NULL;
	int rv					= 0;
	

	PropertySize = SourceImage->GetPropertyItemSize( WhichTag );

	if ( PropertySize ) {
		buffer = (PropertyItem*) malloc( PropertySize );

		if ( SourceImage->GetPropertyItem( WhichTag, PropertySize, buffer ) == 0 ) {
			if ( buffer->type == PropertyTagTypeASCII ) {

				if ( !MultiByteToWideChar( CP_UTF8, MB_PRECOMPOSED | MB_ERR_INVALID_CHARS, (LPCSTR) buffer->value, -1, outString, MAX_ARG_LEN ) ) {
					outString[0] = '\0';
					MultiByteToWideChar( CP_ACP, MB_PRECOMPOSED, (LPCSTR) buffer->value, -1, outString, MAX_ARG_LEN );
				}

				rv = 1;
			}
		}

		free( buffer );
	}

	if ( !rv )
		wcscpy_s( outString, MAX_ARG_LEN, II_NotAvailable );

	return rv;
}



int LooksLikeExifDateTimeString( LPTSTR inString )
{
	if ( !inString )
		return 0;

	if ( !iswdigit( inString[0] ) || !iswdigit( inString[1] ) || !iswdigit( inString[2] ) || !iswdigit( inString[3] ) )
		return 0;

	if ( ( inString[4] != ':' ) || ( inString[7] != ':' ) || ( inString[10] != 0x20 ) || ( inString[13] != ':' ) || ( inString[16] != ':' ) || ( inString[19] != '\0' ) )
		return 0;

	if ( !iswdigit( inString[5] ) || !iswdigit( inString[6] ) || !iswdigit( inString[8] ) || !iswdigit( inString[9] ) )
		return 0;

	if ( !iswdigit( inString[11] ) || !iswdigit( inString[12] ) || !iswdigit( inString[14] ) || !iswdigit( inString[15] ) )
		return 0;

	if ( !iswdigit( inString[17] ) || !iswdigit( inString[18] ) )
		return 0;

	return 1;
}



int ImageFormatToString( GUID ImageFormat, LPTSTR outString )
{
	if ( !outString )
		return -1;


	if ( ImageFormat == ImageFormatUndefined )
		wcscpy_s( outString, MAX_ARG_LEN, L"Undefined" );
	else if ( ImageFormat == ImageFormatMemoryBMP )
		wcscpy_s( outString, MAX_ARG_LEN, L"MemoryBMP" );
	else if ( ImageFormat == ImageFormatBMP )
		wcscpy_s( outString, MAX_ARG_LEN, L"BMP" );
	else if ( ImageFormat == ImageFormatEMF )
		wcscpy_s( outString, MAX_ARG_LEN, L"EMF" );
	else if ( ImageFormat == ImageFormatWMF )
		wcscpy_s( outString, MAX_ARG_LEN, L"WMF" );
	else if ( ImageFormat == ImageFormatJPEG )
		wcscpy_s( outString, MAX_ARG_LEN, L"JPEG" );
	else if ( ImageFormat == ImageFormatPNG )
		wcscpy_s( outString, MAX_ARG_LEN, L"PNG" );
	else if ( ImageFormat == ImageFormatGIF )
		wcscpy_s( outString, MAX_ARG_LEN, L"GIF" );
	else if ( ImageFormat == ImageFormatTIFF )
		wcscpy_s( outString, MAX_ARG_LEN, L"TIFF" );
	else if ( ImageFormat == ImageFormatEXIF )
		wcscpy_s( outString, MAX_ARG_LEN, L"EXIF" );
	else if ( ImageFormat == ImageFormatIcon )
		wcscpy_s( outString, MAX_ARG_LEN, L"Icon" );
	else
		wcscpy_s( outString, MAX_ARG_LEN, L"Unknown" );

	return 0;
}



int BitmapFromClip( Bitmap **OutBitmap )
{
	if ( !OutBitmap )
		return -1;

	HBITMAP ClipboardBitmap			= NULL;
	HPALETTE ClipboardPalette		= NULL;
	Bitmap *NewBitmap				= NULL;


	if ( !OpenClipboard( NULL ) ) {
		*OutBitmap = NULL;
		return 1;
	}

	ClipboardBitmap  = (HBITMAP)  GetClipboardData( CF_BITMAP );
	ClipboardPalette = (HPALETTE) GetClipboardData( CF_PALETTE );

	if ( !ClipboardBitmap ) {
		CloseClipboard();
		*OutBitmap = NULL;
		return 1;
	}

	NewBitmap = new Bitmap( ClipboardBitmap, ClipboardPalette );

	if ( !NewBitmap ) {
		*OutBitmap = NULL;
		return 1;
	}
	else if ( NewBitmap->GetLastStatus() ) {
		delete NewBitmap;
		*OutBitmap = NULL;
		return 1;
	}

	CloseClipboard();
	*OutBitmap = NewBitmap;
	return 0;
}



int ImageFromClip( Image **OutImage )
{
	if ( !OutImage )
		return -1;


	Bitmap *TestBitmap		= NULL;
	Image  *TestImage		= NULL;

	if ( BitmapFromClip( &TestBitmap ) )
		return 1;

	TestImage = dynamic_cast <Image*> ( TestBitmap );
	if ( !TestImage ) {
		*OutImage = NULL;
		return 1;
	}

	*OutImage = TestImage;
	return 0;
}



int OpenImageFile( LPCTSTR inString, Image **outImage, BOOL Complain )
{
	if ( !inString )
		return -1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	Image *SourceImage						= NULL;
	int rv									= 0;


	FH_Truename( inString, Filename );

	
	if ( !_wcsicmp( Filename, L"clip:" ) )
		ImageFromClip( &SourceImage );
	else {
		if ( !QueryIsFile( Filename ) ) {
			if ( Complain ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"File not found: \"%s\"\r\n", Filename );
			}

			return 2;
		}

		SourceImage = new Image( Filename );
	}

	if ( !SourceImage ) {
		if ( Complain ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening file \"%s\"\r\n", Filename );
		}

		return 2;
	}


	rv = SourceImage->GetLastStatus();

	if ( rv ) {
		if ( Complain ) {
			DisplayErrorHeader();

			if ( rv == OutOfMemory )
				Qprintf( ERRHANDLE, L"Not an image file: \"%s\"\r\n", Filename );
			else if ( rv < NumGdiPlusStatusStrings )
				Qprintf( ERRHANDLE, L"GDI+ '%s' error opening file \"%s\"\r\n", GdiPlusStatusStrings[rv], Filename );
			else {
				Qprintf( ERRHANDLE, L"GDI+ error %d opening file \"%s\"\r\n", rv, Filename );
			}
		}

		delete SourceImage;
		return 2;
	}

	*outImage = SourceImage;
	return 0;
}



#ifdef IMGINFO_FUNCTION

DLLExports int WINAPI f_imginfo( LPTSTR String )
{
	if ( !String )
		return 1;


	Image *SourceImage			= NULL;
	unsigned int Which			= 0;
	wchar_t *sep1				= String;
	int t						= 0;
	GUID ImgFormat;


	TrimLine( String );

	if ( iswdigit( String[0] ) ) {
		if ( ParseInt( String, &Which, &sep1 ) == 0 ) {
			if ( ( *sep1 == ',' ) || ( *sep1 == ';' ) ) {
				sep1++;

				while ( iswspace( *sep1 ) )
					sep1++;
			}
			else {
				Which = 0;
				sep1 = String;
			}
		}
		else {
			Which = 0;
			sep1 = String;
		}
	}


	StripQuotes( sep1 );

	if ( *sep1 == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}


	if ( OpenImageFile( sep1, &SourceImage, TRUE ) )
		return -1;


	switch ( Which ) {

	case 1 :

		_ultow_s( SourceImage->GetWidth(), String, MAX_ARG_LEN, 10 );
		break;

	case 2 :

		_ultow_s( SourceImage->GetHeight(), String, MAX_ARG_LEN, 10 );
		break;

	case 3 :

		_ltow_s( (int) SourceImage->GetHorizontalResolution(), String, MAX_ARG_LEN, 10 );
		break;

	case 4 :

		_ltow_s( (int) SourceImage->GetVerticalResolution(), String, MAX_ARG_LEN, 10 );
		break;

	case 5 :

		t = SourceImage->GetPixelFormat();
		_ultow_s( ( t & 0xff00 ) >> 8, String, MAX_ARG_LEN, 10 );
		break;

	case 6 :

		if ( SourceImage->GetRawFormat( &ImgFormat ) )
			String[0] = '\0';
		else
			ImageFormatToString( ImgFormat, String );

		break;

	case 10 :

		t = GetExifShort( SourceImage, PropertyTagOrientation );

		if ( ( t < 1 ) || ( t > 8 ) )
			wcscpy_s( String, MAX_ARG_LEN, II_NotAvailable );
		else
			_ltow_s( t, String, MAX_ARG_LEN, 10 );

		break;

	case 11 :
	case 12 :
	case 13 :

		{
			PROPID WhichStamp		= PropertyTagDateTime;


			if ( Which == 12 )
				WhichStamp = PropertyTagExifDTOrig;
			else if ( Which == 13 )
				WhichStamp = PropertyTagExifDTDigitized;

			GetExifString( SourceImage, WhichStamp, String );

			if ( LooksLikeExifDateTimeString( String ) ) {
				String[4] = '-';
				String[7] = '-';
			}
			else
				wcscpy_s( String, MAX_ARG_LEN, II_NotAvailable );
		}
		break;

	case 14 :

		GetExifString( SourceImage, PropertyTagArtist, String );
		break;

	case 15 :

		GetExifString( SourceImage, PropertyTagImageDescription, String );
		break;

	case 16 :

		GetExifString( SourceImage, PropertyTagCopyright, String );
		break;

	case 17 :

		GetExifString( SourceImage, PropertyTagSoftwareUsed, String );
		break;

	case 18 :

		GetExifString( SourceImage, PropertyTagEquipMake, String );
		break;

	case 19 :

		GetExifString( SourceImage, PropertyTagEquipModel, String );
		break;

	default :

		wsprintf( String, L"%u \u00d7 %u", SourceImage->GetWidth(), SourceImage->GetHeight() );
	}

	delete( SourceImage );
	return 0;
}

#endif	//	IMGINFO_FUNCTION

#endif	//	IMGINFO_VER


/*

	1.0.0.0		2022-03-10

	1.0.0.1		2022-03-17
				Specifically check for file not found.

	1.0.0.2		2023-08-22
				ImageFileType() and OpenImageFile() filename args defined as constant.

	1.0.0.3		2023-12-14
				Made f_imginfo() optional; #define IMGINFO_FUNCTION to enable.

*/
