
//		RPTarget.cpp -- resolve reparse point targets


#ifndef RPTARGET_VER
#define RPTARGET_VER			1.0.0.1


#include "ReparseDataBuffer.h"



#ifdef RPTARGET_HELP_TEXT

const wchar_t F_RPTargetHelpText[]	= L"@RPTARGET[filename] \r\n\
\n\
Returns the target of a reparse point.  \n";

#endif	//	RPTARGET_HELP_TEXT



int GetReparsePointTarget( LPTSTR inFilename, LPTSTR outTarget, size_t buflen )
{
	if ( !inFilename || !outTarget || ( buflen < MAX_PATH ) )
		return -1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";

	_REPARSE_DATA_BUFFER *ReparseData;

	HANDLE FileHandle						= NULL;
	DWORD BytesReturned						= 0;
	size_t NameOffset						= 0;
	size_t NameLen							= 0;

	int rv									= -1;


	wcscpy_s( Filename, MAX_FILENAME_LEN, inFilename );


	ReparseData = (_REPARSE_DATA_BUFFER*) malloc( MAXIMUM_REPARSE_DATA_BUFFER_SIZE );
	if ( !ReparseData )
		return ERROR_NOT_ENOUGH_MEMORY;


	FileHandle = CreateFile( Filename, 0, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT, NULL );
	if ( !FileHandle || ( FileHandle == INVALID_HANDLE_VALUE ) ) {
		rv = GetLastError();
		free( ReparseData );
		return rv;
	}


	if ( DeviceIoControl( FileHandle, FSCTL_GET_REPARSE_POINT, NULL, 0, ReparseData, MAXIMUM_REPARSE_DATA_BUFFER_SIZE, &BytesReturned, NULL ) ) {

		if ( ReparseData->ReparseTag == IO_REPARSE_TAG_SYMLINK ) {
			NameOffset	= ReparseData->SymbolicLinkReparseBuffer.SubstituteNameOffset / sizeof( wchar_t );
			NameLen		= ReparseData->SymbolicLinkReparseBuffer.SubstituteNameLength / sizeof( wchar_t );

			if ( ReparseData->SymbolicLinkReparseBuffer.Flags & 1 ) {
				//	relative symlink

				wchar_t Path[MAX_FILENAME_LEN]			= L"";

				FH_PathPart( Filename, Path );
				if ( Path[0] == '\0' )
					GetCurrentDirectory( MAX_FILENAME_LEN, Path );

				GetFullPathName( Path, (DWORD) buflen, outTarget, NULL );		//	QueryTrueName( Path, String );
				FH_AppendFilename( outTarget, L"", buflen );
				wcsncat_s( outTarget, buflen, ReparseData->SymbolicLinkReparseBuffer.PathBuffer + NameOffset, NameLen );
				rv = 0;
			}
			else {
				//	absolute symlink

				if ( !wcsncmp( ReparseData->SymbolicLinkReparseBuffer.PathBuffer + NameOffset, L"\\??\\", 4 ) )
					wcsncpy_s( outTarget, buflen, ReparseData->SymbolicLinkReparseBuffer.PathBuffer + NameOffset + 4, NameLen - 4 );
				else
					wcsncpy_s( outTarget, buflen, ReparseData->SymbolicLinkReparseBuffer.PathBuffer + NameOffset, NameLen );

				rv = 0;
			}
		}
		else if ( ReparseData->ReparseTag == IO_REPARSE_TAG_MOUNT_POINT ) {
			NameOffset	= ReparseData->MountPointReparseBuffer.SubstituteNameOffset / sizeof( wchar_t );
			NameLen		= ReparseData->MountPointReparseBuffer.SubstituteNameLength / sizeof( wchar_t );

			if ( ( wcsncmp( ReparseData->MountPointReparseBuffer.PathBuffer + NameOffset, L"\\??\\", 4 ) == 0 ) && ReparseData->MountPointReparseBuffer.PathBuffer[NameOffset + 4] ) {
				wcsncpy_s( outTarget, buflen, ReparseData->MountPointReparseBuffer.PathBuffer + NameOffset + 4, NameLen - 4 );
				outTarget[NameLen - 3] = '\0';
				rv = 0;
			}
		}
		else if ( ReparseData->ReparseTag == IO_REPARSE_TAG_APPX_ALIAS__ ) {
			//	app alias

			LPTSTR PackageID	= ReparseData->AppXAlias.StringList;
			LPTSTR EntryPoint	= PackageID + wcslen( PackageID ) + 1;
			LPTSTR Executable	= EntryPoint + wcslen( EntryPoint ) + 1;
			LPTSTR ApplicType	= Executable + wcslen( Executable ) + 1;

			/*
			wprintf( L"   *  ReparseTag = 0x%08X    ReparseDataLength = %u    Version = %u\n", ReparseData->ReparseTag, ReparseData->ReparseDataLength, ReparseData->AppXAlias.Version );
			wprintf( L"   *  PackageID  = %s\n", PackageID );
			wprintf( L"   *  EntryPoint = %s\n", EntryPoint );
			wprintf( L"   *  Executable = %s\n", Executable );
			wprintf( L"   *  ApplicType = %s\n", ApplicType );
			*/

			wcscpy_s( outTarget, buflen, Executable );
			rv = 0;
		}
		else {
			wsprintf( outTarget, L"ReparseData-\u2192ReparseTag = 0x%08X\n", ReparseData->ReparseTag );
			rv = ERROR_INVALID_DATA;
		}
	}
	else
		rv = GetLastError();


	CloseHandle( FileHandle );
	free( ReparseData );
	return rv;
}



DLLExports int WINAPI f_rptarget( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t ErrorText[SHORT_BUF_LEN]			= L"";

	int rv				= 0;



	TrimLine( String );
	StripQuotes( String );


	if ( String[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}

	if ( rv = GetReparsePointTarget( String, String, MAX_ARG_LEN ) ) {
		GetErrorText( rv, ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error resolving \"%s\" :\r\n", String );
		Qprintf( ERRHANDLE, L"    %s\r\n", ErrorText );

		if ( rv == ERROR_INVALID_DATA )
			Qprintf( ERRHANDLE, L"    %s\r\n", String );

		return -1;
	}

	return 0;
}


#endif	//	RPTARGET_VER


/*

	1.0.0.0		2022-12-19

	1.0.0.1		2023-08-04
				Adds optional help text; #define RPTARGET_HELP_TEXT to enable.

*/
