
#ifndef WINPROP_VER
#define WINPROP_VER				1.0.2.0


#include <propkey.h>
#include <propvarutil.h>




int GetPropString( IPropertyStore *Store, REFPROPERTYKEY Key, LPTSTR String, UINT len )
{
	if ( !Store || !String || !len )
		return -1;


	PROPVARIANT Var;


	memset( &Var, 0x00, sizeof( PROPVARIANT ) );
	String[0] = '\0';

	if ( !SUCCEEDED( Store->GetValue( Key, &Var ) ) )
		return 1;

	if ( PropVariantToString( Var, String, len ) ) {
		String[0] = '\0';
		return 2;
	}

	return 0;
}



int GetPropUnsigned( IPropertyStore *Store, REFPROPERTYKEY Key, unsigned int *outInt )
{
	if ( !Store || !outInt )
		return -1;


	PROPVARIANT Var;


	memset( &Var, 0x00, sizeof( PROPVARIANT ) );
	*outInt = 0;

	if ( !SUCCEEDED( Store->GetValue( Key, &Var ) ) )
		return 1;

	if ( Var.vt != VT_UI4 )
		return 2;

	*outInt = Var.uintVal;
	return 0;
}



int GetPropUnsigned64( IPropertyStore *Store, REFPROPERTYKEY Key, unsigned __int64 *outInt )
{
	if ( !Store || !outInt )
		return -1;


	PROPVARIANT Var;


	memset( &Var, 0x00, sizeof( PROPVARIANT ) );
	*outInt = 0;

	if ( !SUCCEEDED( Store->GetValue( Key, &Var ) ) )
		return 1;

	if ( Var.vt != VT_UI8 )
		return 2;

	*outInt = Var.uhVal.QuadPart;
	return 0;
}



int GetPropFileTime( IPropertyStore *Store, REFPROPERTYKEY Key, LPTSTR String )
{
	if ( !Store || !String )
		return -1;


	PROPVARIANT Var;
	SYSTEMTIME SysTime;


	memset( &Var, 0x00, sizeof( PROPVARIANT ) );
	memset( &SysTime, 0x00, sizeof( SYSTEMTIME ) );
	String[0] = '\0';

	if ( !SUCCEEDED( Store->GetValue( Key, &Var ) ) )
		return 1;

	if ( Var.vt != VT_FILETIME )
		return 2;

	if ( !FileTimeToSystemTime( &( Var.filetime ), &SysTime ) )
		return 3;

	wsprintf( String, L"%04u-%02u-%02u %02u:%02u", SysTime.wYear, SysTime.wMonth, SysTime.wDay, SysTime.wHour, SysTime.wMinute );
	return 0;
}



DLLExports int WINAPI f_winprop( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t Filename[MAX_FILENAME_LEN]		= L"";
	wchar_t temp[SHORT_BUF_LEN]				= L"";
	IPropertyStore *Store					= NULL;

	unsigned int Which			= 0;
	unsigned int UInt			= 0;
	unsigned __int64 UInt64		= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[0] ) {
		if ( ParseInt( arg[0], &Which, NULL ) ) {
			_wcsupr_s( arg[0], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", String );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing value\r\n" );
		return -1;
	}


	if ( arglen[1] ) {
		FH_Truename( arg[1], Filename );

		if ( Filename[0] == '\0' ) {
			_wcsupr_s( arg[1], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad filename: \"%s\"\r\n", arg[1] );
			return -1;
		}

		if ( !QueryIsFile( Filename ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File not found: \"%s\"\r\n", Filename );
			return -1;
		}
	}
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return -1;
	}


	String[0] = '\0';

	SHGetPropertyStoreFromParsingName( Filename, NULL, GPS_DEFAULT, __uuidof( IPropertyStore ), (void**) &Store );
	if ( !Store )
		goto Quit;


	switch ( Which ) {

	case 0 :

		GetPropString( Store, PKEY_Author, String, MAX_ARG_LEN );
		break;

	case 1 :

		GetPropString( Store, PKEY_Comment, String, MAX_ARG_LEN );
		break;

	case 2 :

		GetPropString( Store, PKEY_Copyright, String, MAX_ARG_LEN );
		break;

	case 3 :

		GetPropString( Store, PKEY_FileDescription, String, MAX_ARG_LEN );
		break;

	case 4 :
		
		GetPropString( Store, PKEY_FileVersion, String, MAX_ARG_LEN );
		break;

	case 5 :

		GetPropString( Store, PKEY_ItemAuthors, String, MAX_ARG_LEN );
		break;

	case 6 :

		GetPropString( Store, PKEY_ItemNameDisplay, String, MAX_ARG_LEN );
		break;

	case 7 :

		GetPropString( Store, PKEY_Keywords, String, MAX_ARG_LEN );
		break;

	case 8 :

		if ( GetPropUnsigned( Store, PKEY_Rating, &UInt ) )
			;
		else if ( UInt < RATING_ONE_STAR_MIN )
			;
		else if ( UInt <= RATING_ONE_STAR_MAX )
			wcscpy_s( String, MAX_ARG_LEN, L"*" );
		else if ( UInt <= RATING_TWO_STARS_MAX )
			wcscpy_s( String, MAX_ARG_LEN, L"* *" );
		else if ( UInt <= RATING_THREE_STARS_MAX )
			wcscpy_s( String, MAX_ARG_LEN, L"* * *" );
		else if ( UInt <= RATING_FOUR_STARS_MAX )
			wcscpy_s( String, MAX_ARG_LEN, L"* * * *" );
		else if ( UInt <= RATING_FIVE_STARS_MAX )
			wcscpy_s( String, MAX_ARG_LEN, L"* * * * *" );

		break;

	case 9 :

		GetPropString( Store, PKEY_Subject, String, MAX_ARG_LEN );
		break;

	case 10 :

		GetPropString( Store, PKEY_Title, String, MAX_ARG_LEN );
		break;

	case 11 :

		GetPropString( Store, PKEY_Trademarks, String, MAX_ARG_LEN );
		break;

	case 20 :

		GetPropString( Store, PKEY_Media_AuthorUrl, String, MAX_ARG_LEN );
		break;

	case 21 :

		GetPropString( Store, PKEY_Media_CreatorApplication, String, MAX_ARG_LEN );
		break;

	case 22 :

		GetPropString( Store, PKEY_Media_CreatorApplicationVersion, String, MAX_ARG_LEN );
		break;

	case 23 :

		GetPropFileTime( Store, PKEY_Media_DateEncoded, String );
		break;

	case 24 :

		GetPropString( Store, PKEY_Media_DateReleased, String, MAX_ARG_LEN );
		break;

	case 25 :

		if ( !GetPropUnsigned64( Store, PKEY_Media_Duration, &UInt64 ) ) {
			unsigned __int64 sec		= UInt64 / 10000000;
			unsigned __int64 min		= 0;
			unsigned __int64 hrs		= 0;


			if ( sec >= 60 )
				min = sec / 60, sec %= 60;

			if ( min >= 60 )
				hrs = min / 60, min %= 60;

			wsprintf( String, L"%I64u:%02I64u:%02I64u", hrs, min, sec );
		}

		break;

	case 26 :

		GetPropString( Store, PKEY_Media_Producer, String, MAX_ARG_LEN );
		break;

	case 27 :

		GetPropString( Store, PKEY_Media_Publisher, String, MAX_ARG_LEN );
		break;

	case 28 :

		GetPropString( Store, PKEY_Media_SubTitle, String, MAX_ARG_LEN );
		break;

	case 29 :

		if ( !GetPropUnsigned( Store, PKEY_Media_Year, &UInt ) )
			_ultow_s( UInt, String, MAX_ARG_LEN, 10 );

		break;

	case 40 :

		GetPropString( Store, PKEY_Music_AlbumArtist, String, MAX_ARG_LEN );
		break;

	case 41 :

		GetPropString( Store, PKEY_Music_AlbumID, String, MAX_ARG_LEN );
		break;

	case 42 :

		GetPropString( Store, PKEY_Music_AlbumTitle, String, MAX_ARG_LEN );
		break;

	case 43 :

		GetPropString( Store, PKEY_Music_Artist, String, MAX_ARG_LEN );
		break;

	case 44 :

		GetPropString( Store, PKEY_Music_Composer, String, MAX_ARG_LEN );
		break;

	case 45 :

		GetPropString( Store, PKEY_Music_Conductor, String, MAX_ARG_LEN );
		break;

	case 46 :

		GetPropString( Store, PKEY_Music_DisplayArtist, String, MAX_ARG_LEN );
		break;

	case 47 :

		GetPropString( Store, PKEY_Music_Genre, String, MAX_ARG_LEN );
		break;

	case 48 :

		GetPropString( Store, PKEY_Music_Mood, String, MAX_ARG_LEN );
		break;

	case 49 :

		GetPropString( Store, PKEY_Music_Period, String, MAX_ARG_LEN );
		break;

	case 50 :

		if ( !GetPropUnsigned( Store, PKEY_Music_TrackNumber, &UInt ) )
			_ultow_s( UInt, String, MAX_ARG_LEN, 10 );

		break;

	//		Audio properties :

	case 60 :

		if ( !GetPropUnsigned( Store, PKEY_Audio_ChannelCount, &UInt ) )
			_ultow_s( UInt, String, MAX_ARG_LEN, 10 );

		break;

	case 61 :

		GetPropString( Store, PKEY_Audio_Compression, String, MAX_ARG_LEN );
		break;

	case 62 :

		if ( !GetPropUnsigned( Store, PKEY_Audio_EncodingBitrate, &UInt ) )
			_ultow_s( UInt, String, MAX_ARG_LEN, 10 );

		break;

	case 63 :

		GetPropString( Store, PKEY_Audio_Format, String, MAX_ARG_LEN );
		break;

	case 65 :

		if ( !GetPropUnsigned( Store, PKEY_Audio_PeakValue, &UInt ) )
			_ultow_s( UInt, String, MAX_ARG_LEN, 10 );

		break;

	case 66 :

		if ( !GetPropUnsigned( Store, PKEY_Audio_SampleRate, &UInt ) )
			_ultow_s( UInt, String, MAX_ARG_LEN, 10 );

		break;

	case 67 :

		if ( !GetPropUnsigned( Store, PKEY_Audio_SampleSize, &UInt ) )
			_ultow_s( UInt, String, MAX_ARG_LEN, 10 );

		break;


	default:

		String[0] = '\0';
	}


Quit:

	if ( Store )
		Store->Release();

	return 0;
}


#endif	//	WINPROP_VER


/*

	1.0.0.0		2022-03-10

	1.0.1.0		2022-03-17
				Added 50 :  System.Music.TrackNumber.

	1.0.2.0		2022-06-09
				Added:	60 :  System.Audio.ChannelCount
						61 :  System.Audio.Compression
						62 :  System.Audio.EncodingBitrate
						63 :  System.Audio.Format
						65 :  System.Audio.PeakValue
						66 :  System.Audio.SampleRate
						67 :  System.Audio.SampleSize

*/
